"""The n-dimensional hyperbolic space.

The n-dimensional hyperbolic space embedded and its different representations.
"""


import geomstats.backend as gs
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric

TOLERANCE = 1e-6


class Hyperbolic(Manifold):
    """Class for the n-dimensional hyperbolic space.

    Class for the n-dimensional hyperbolic space
    as embedded in (n+1)-dimensional Minkowski space.

    The point_type variable allows to choose the
    representation of the points as input.

    If point_type is set to 'ball' then points are parametrized
    by their coordinates inside the Poincare Ball n-coordinates.

    Parameters
    ----------
    dim : int
        Dimension of the hyperbolic space.
    point_type : str, {'extrinsic', 'intrinsic', etc}, optional
        Default coordinates to represent points in hyperbolic space.
    scale : int, optional
        Scale of the hyperbolic space, defined as the set of points
        in Minkowski space whose squared norm is equal to -scale.
    """

    default_coords_type = 'extrinsic'
    default_point_type = 'vector'

    def __init__(self, dim, scale=1):
        super(Hyperbolic, self).__init__(dim=dim)
        self.point_type = Hyperbolic.default_point_type
        self.coords_type = Hyperbolic.default_coords_type
        self.scale = scale
        self.metric = HyperbolicMetric(self.dim, self.scale)

    @staticmethod
    def _extrinsic_to_extrinsic_coordinates(point):
        """Convert the parameterization of a point.

        Convert the parameterization of a point in hyperbolic space
        from its extrinsic coordinates, to its intrinsic coordinates
        in Minkowski space.

        Parameters
        ----------
        point_extrinsic : array-like, shape=[n_samples, dim + 1]
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        point_intrinsic : array-like, shape=[n_samples, dim]
            Point in hyperbolic space in intrinsic coordinates.
        """
        return gs.to_ndarray(point, to_ndim=2)

    @staticmethod
    def _intrinsic_to_extrinsic_coordinates(point_intrinsic):
        """Convert intrinsic to extrinsic coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its intrinsic coordinates, to its extrinsic coordinates
        in Minkowski space.

        Parameters
        ----------
        point_intrinsic : array-like, shape=[n_samples, dim]
            Point in hyperbolic space in intrinsic coordinates.

        Returns
        -------
        point_extrinsic : array-like, shape=[n_samples, dim + 1]
            Point in hyperbolic space in extrinsic coordinates.
        """
        point_intrinsic = gs.to_ndarray(point_intrinsic, to_ndim=2)

        coord_0 = gs.sqrt(1. + gs.linalg.norm(point_intrinsic, axis=-1) ** 2)
        coord_0 = gs.to_ndarray(coord_0, to_ndim=2, axis=1)

        point_extrinsic = gs.concatenate([coord_0, point_intrinsic], axis=-1)

        return point_extrinsic

    @staticmethod
    def _extrinsic_to_intrinsic_coordinates(point_extrinsic):
        """Convert extrinsic to intrinsic coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its extrinsic coordinates in Minkowski space, to its
        intrinsic coordinates.

        Parameters
        ----------
        point_extrinsic : array-like, shape=[n_samples, dim + 1]
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        point_intrinsic : array-like, shape=[n_samples, dim]
        """
        point_extrinsic = gs.to_ndarray(point_extrinsic, to_ndim=2)

        point_intrinsic = point_extrinsic[:, 1:]

        return point_intrinsic

    @staticmethod
    def _extrinsic_to_ball_coordinates(point):
        """Convert extrinsic to ball coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its intrinsic coordinates, to the poincare ball model
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim + 1]
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        point_ball : array-like, shape=[n_samples, dim]
            Point in hyperbolic space in Poincare ball coordinates.
        """
        return point[:, 1:] / (1 + point[:, :1])

    @staticmethod
    def _ball_to_extrinsic_coordinates(point):
        """Convert ball to extrinsic coordinates.

        Convert the parameterization of a point in hyperbolic space
        from its poincare ball model coordinates, to the extrinsic
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
            Point in hyperbolic space in Poincare ball coordinates.

        Returns
        -------
        extrinsic : array-like, shape=[n_samples, dim + 1]
            Point in hyperbolic space in extrinsic coordinates.
        """
        squared_norm = gs.sum(point**2, -1)
        denominator = 1 - squared_norm
        t = gs.to_ndarray((1 + squared_norm) / denominator, to_ndim=2, axis=1)
        expanded_denominator = gs.expand_dims(denominator, -1)
        expanded_denominator = gs.repeat(
            expanded_denominator, point.shape[-1], -1)
        intrinsic = (2 * point) / expanded_denominator
        return gs.concatenate([t, intrinsic], -1)

    @staticmethod
    def _half_plane_to_extrinsic_coordinates(point):
        """Convert half plane to extrinsic coordinates.

        Convert the parameterization of a point in the hyperbolic plane
        from its upper half plane model coordinates, to the extrinsic
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[n_samples, 2]
            Point in hyperbolic space in half-plane coordinates.

        Returns
        -------
        extrinsic : array-like, shape=[n_samples, 3]
            Point in hyperbolic plane in extrinsic coordinates.
        """
        x, y = point[:, 0], point[:, 1]
        x2 = point[:, 0]**2
        den = x2 + (1 + y)**2
        x = gs.to_ndarray(x, to_ndim=2, axis=0)
        y = gs.to_ndarray(y, to_ndim=2, axis=0)
        x2 = gs.to_ndarray(x2, to_ndim=2, axis=0)
        den = gs.to_ndarray(den, to_ndim=2, axis=0)
        ball_point = gs.hstack((2 * x / den, (x2 + y**2 - 1) / den))
        return Hyperbolic._ball_to_extrinsic_coordinates(ball_point)

    @staticmethod
    def _extrinsic_to_half_plane_coordinates(point):
        """Convert extrinsic to half plane coordinates.

        Convert the parameterization of a point in the hyperbolic plane
        from its intrinsic coordinates, to the poincare upper half plane
        coordinates.

        Parameters
        ----------
        point : array-like, shape=[n_samples, 2]
            Point in the hyperbolic plane in intrinsic coordinates.

        Returns
        -------
        point_half_plane : array-like, shape=[n_samples, 2]
            Point in the hyperbolic plane in Poincare upper half-plane
            coordinates.
        """
        point_ball = \
            Hyperbolic._extrinsic_to_ball_coordinates(point)
        point_ball_x, point_ball_y = point_ball[:, 0], point_ball[:, 1]
        point_ball_x2 = point_ball_x**2
        denom = point_ball_x2 + (1 - point_ball_y)**2

        point_ball_x = gs.to_ndarray(
            point_ball_x, to_ndim=2, axis=0)
        point_ball_y = gs.to_ndarray(
            point_ball_y, to_ndim=2, axis=0)
        point_ball_x2 = gs.to_ndarray(
            point_ball_x2, to_ndim=2, axis=0)
        denom = gs.to_ndarray(
            denom, to_ndim=2, axis=0)

        point_half_plane = gs.hstack((
            (2 * point_ball_x) / denom,
            (1 - point_ball_x2 - point_ball_y**2) / denom))
        return point_half_plane

    @staticmethod
    def change_coordinates_system(point,
                                  from_coordinates_system,
                                  to_coordinates_system):
        """Convert coordinates of a point.

        Convert the parameterization of a point in the hyperbolic space
        from current given coordinate system to an other also given in
        parameters. The possible coordinates system are 'extrinsic',
        'intrinsic', 'ball' and 'half-plane' that correspond respectivelly
        to extrinsic coordinates in the hyperboloid , intrinsic
        coordinates in the hyperboloid, ball coordinates in the Poincare
        ball model and coordinates in the Poincare upper half-plane model.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
                            or shape=[n_samples, dim + 1]
            Point in hyperbolic space.
        from_coordinates_system : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.
        to_coordinates_system : str, {'extrinsic', 'intrinsic', etc}
            Coordinates type.

        Returns
        -------
        point_to : array-like, shape=[n_samples, dim]
                               or shape=[n_sample, dim + 1]
            Point in hyperbolic space in coordinates given by to_point_type.
        """
        coords_transform = {
            'ball-extrinsic':
                Hyperbolic._ball_to_extrinsic_coordinates,
            'extrinsic-ball':
                Hyperbolic._extrinsic_to_ball_coordinates,
            'intrinsic-extrinsic':
                Hyperbolic._intrinsic_to_extrinsic_coordinates,
            'extrinsic-intrinsic':
                Hyperbolic._extrinsic_to_intrinsic_coordinates,
            'extrinsic-half-plane':
                Hyperbolic._extrinsic_to_half_plane_coordinates,
            'half-plane-extrinsic':
                Hyperbolic._half_plane_to_extrinsic_coordinates,
            'extrinsic-extrinsic':
                Hyperbolic._extrinsic_to_extrinsic_coordinates
        }

        point = gs.to_ndarray(point, to_ndim=2, axis=0)

        if from_coordinates_system == to_coordinates_system:
            return point

        extrinsic =\
            coords_transform[from_coordinates_system +
                             '-extrinsic'](point)
        return \
            coords_transform['extrinsic-' +
                             to_coordinates_system](extrinsic)

    def belongs(self, point, tolerance=TOLERANCE):
        """Test if a point belongs to the hyperbolic space.

        Test if a point belongs to the hyperbolic space in
        its hyperboloid representation.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
            Point to be tested.
        tolerance : float, optional
            Tolerance at which to evaluate how close the squared norm
            is to the reference value.

        Returns
        -------
        belongs : array-like, shape=[n_samples, 1]
            Array of booleans indicating whether the corresponding points
            belong to the hyperbolic space.
        """
        raise NotImplementedError

    def to_coordinates(self, point, to_coords_type='ball'):
        """Convert coordinates of a point.

        Convert the parameterization of a point in the hyperbolic space
        from current coordinate system to the coordinate system given.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
                            or shape=[n_samples, dim + 1]
            Point in hyperbolic space.
        to_point_type : str, {'extrinsic', 'intrinsic', etc}, optional
            Coordinates type.

        Returns
        -------
        point_to : array-like, shape=[n_samples, dim]
                               or shape=[n_sample, dim + 1]
            Point in hyperbolic space in coordinates given by to_point_type.
        """
        return Hyperbolic.change_coordinates_system(point,
                                                    self.coords_type,
                                                    to_coords_type)

    def from_coordinates(self, point, from_coords_type):
        """Convert to a type of coordinates given some type.

        Convert the parameterization of a point in hyperbolic space
        from given coordinate system to the current coordinate system.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
                            or shape=[n_samples, dim + 1]
            Point in hyperbolic space in coordinates from_point_type.
        from_point_type : str, {'ball', 'extrinsic', 'intrinsic', 'half_plane'}
            Coordinates type.

        Returns
        -------
        point_current : array-like, shape=[n_samples, dim]
                                    or shape=[n_sample, dim + 1]
            Point in hyperbolic space.
        """
        return Hyperbolic.change_coordinates_system(point,
                                                    from_coords_type,
                                                    self.coords_type)

    def random_uniform(self, n_samples=1, bound=1.):
        """Sample over the hyperbolic space using uniform distribution.

        Sample over the hyperbolic space. The sampling is performed
        by sampling over uniform distribution, the sampled examples
        are considered in the intrinsic coordinates system.
        The function then transforms intrinsic samples into system
        coordinate selected.

        Parameters
        ----------
        n_samples : int, optional
            Number of samples.
        bound: float, optional
            Bound defining the hypersquare in which to sample uniformly.

        Returns
        -------
        samples : array-like, shape=[n_samples, dim + 1]
            Samples in hyperbolic space.
        """
        size = (n_samples, self.dim)
        samples = bound * 2. * (gs.random.rand(*size) - 0.5)

        samples = Hyperbolic.change_coordinates_system(
            samples, 'intrinsic', self.coords_type)

        if n_samples == 1:
            samples = gs.squeeze(samples, axis=0)
        return samples


class HyperbolicMetric(RiemannianMetric):
    """Class that defines operations using a hyperbolic metric.

    Parameters
    ----------
    dim : int
        Dimension of the hyperbolic space.
    point_type : str, {'extrinsic', 'intrinsic', etc}, optional
        Default coordinates to represent points in hyperbolic space.
    scale : int, optional
        Scale of the hyperbolic space, defined as the set of points
        in Minkowski space whose squared norm is equal to -scale.
    """

    default_point_type = 'vector'
    default_coords_type = 'extrinsic'

    def __init__(self, dim, scale=1):
        super(HyperbolicMetric, self).__init__(
            dim=dim,
            signature=(dim, 0, 0))
        self.point_type = HyperbolicMetric.default_point_type

        self.scale = scale

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute the inner-product of two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[n_samples, dim + 1]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[n_samples, dim + 1]
            Second tangent vector at base point.
        base_point : array-like, shape=[n_samples, dim + 1], optional
            Point in hyperbolic space.

        Returns
        -------
        inner_prod : array-like, shape=[n_samples, 1]
            Inner-product of the two tangent vectors.
        """
        inner_prod = self._inner_product(
            tangent_vec_a, tangent_vec_b, base_point)
        inner_prod *= self.scale ** 2
        return inner_prod

    def squared_norm(self, vector, base_point=None):
        """Compute the squared norm of a vector.

        Squared norm of a vector associated with the inner-product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[n_samples, dim + 1]
            Vector on the tangent space of the hyperbolic space at base point.
        base_point : array-like, shape=[n_samples, dim + 1], optional
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        sq_norm : array-like, shape=[n_samples, 1]
            Squared norm of the vector.
        """
        sq_norm = self._squared_norm(vector)
        sq_norm *= self.scale ** 2
        return sq_norm

    def _squared_norm(self, vector, base_point=None):
        """Squared norm with hyperbolic scale 1.

        Squared norm of a vector associated with the inner-product
        at the tangent space at a base point with scale=1.

        Parameters
        ----------
        vector : array-like, shape=[n_samples, dim + 1]
            Vector on the tangent space of the hyperbolic space at base point.
        base_point : array-like, shape=[n_samples, dim + 1], optional
            Point in hyperbolic space in extrinsic coordinates.

        Returns
        -------
        sq_norm : array-like, shape=[n_samples, 1]
            Squared norm of the vector.
        """
        return super().squared_norm(vector, base_point=base_point)

    def _inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute the inner-product of two tangent vectors with scale=1.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[n_samples, dim + 1]
            First tangent vector at base point.
        tangent_vec_b : array-like, shape=[n_samples, dim + 1]
            Second tangent vector at base point.
        base_point : array-like, shape=[n_samples, dim + 1], optional
            Point in hyperbolic space.

        Returns
        -------
        inner_prod : array-like, shape=[n_samples, 1]
            Inner-product of the two tangent vectors.
        """
        return super().inner_product(tangent_vec_a, tangent_vec_b,
                                     base_point=base_point)



"""The special orthogonal group SO(n).

i.e. the Lie group of rotations in n dimensions.
"""

import geomstats.backend as gs
import geomstats.vectorization
from geomstats import algebra_utils
from geomstats.geometry.embedded_manifold import EmbeddedManifold
from geomstats.geometry.general_linear import GeneralLinear
from geomstats.geometry.lie_group import LieGroup

ATOL = 1e-5

TAYLOR_COEFFS_1_AT_0 = [1., 0.,
                        - 1. / 12., 0.,
                        - 1. / 720., 0.,
                        - 1. / 30240., 0.]
TAYLOR_COEFFS_2_AT_0 = [1. / 12., 0.,
                        1. / 720., 0.,
                        1. / 30240., 0.,
                        1. / 1209600., 0.]
TAYLOR_COEFFS_1_AT_PI = [0., - gs.pi / 4.,
                         - 1. / 4., - gs.pi / 48.,
                         - 1. / 48., - gs.pi / 480.,
                         - 1. / 480.]


class SpecialOrthogonal(LieGroup, EmbeddedManifold):
    """Class for the special orthogonal group SO(n).

    i.e. the Lie group of rotations.
    """

    def __init__(self, n, default_point_type=None, epsilon=0.):
        """Initialize an instance of SO(n).

        Parameters
        ----------
        n : int
            the dimension of the euclidean space that SO(n) acts upon
        default_point_type : str, {'vector', 'matrix'}, optional
            if None is given, point_type is set to 'vector for dimension 3
            and matrix otherwise
        epsilon : float, optional
            precision to use for calculations involving potential divison by in
            rotations
            default: 0
        """
        if not (isinstance(n, int) and n > 1):
            raise ValueError('Parameter `n` is required to be an integer > 1.')

        self.n = n
        self.dim = int((n * (n - 1)) / 2)

        self.epsilon = epsilon
        if default_point_type is None:
            default_point_type = 'vector' if n == 3 else 'matrix'

        LieGroup.__init__(self,
                          dim=self.dim, default_point_type=default_point_type)
        EmbeddedManifold.__init__(self,
                                  dim=self.dim,
                                  embedding_manifold=GeneralLinear(n=n),
                                  default_point_type=default_point_type)
        self.bi_invariant_metric = self.left_canonical_metric

    def get_identity(self, point_type=None):
        """Get the identity of the group.

        Parameters
        ----------
        point_type : str, {'vector', 'matrix'}, optional
            the point_type of the returned value

        Returns
        -------
        identity : array-like, shape={[dim], [n, n]}
        """
        if point_type is None:
            point_type = self.default_point_type

        identity = gs.zeros(self.dim)
        if point_type == 'matrix':
            identity = gs.eye(self.n)
        return identity

    identity = property(get_identity)

    def belongs(self, point, point_type=None):
        """Evaluate if a point belongs to SO(n).

        Parameters
        ----------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
            the point of which to check whether it belongs to SO(n)
        point_type : str, {'vector', 'matrix'}, optional
            default: default_point_type

        Returns
        -------
        belongs : array-like, shape=[n_samples, 1]
            array of booleans indicating whether point belongs to SO(n)
        """
        if point_type is None:
            point_type = self.default_point_type

        if point_type == 'vector':
            vec_dim = point.shape[-1]
            belongs = vec_dim == self.dim
            if point.ndim == 2:
                belongs = gs.tile([belongs], (point.shape[0],))
            return belongs

        if point_type == 'matrix':
            point = gs.to_ndarray(point, to_ndim=3)
            point_transpose = gs.transpose(point, axes=(0, 2, 1))
            mask = gs.isclose(gs.matmul(point, point_transpose),
                              gs.eye(self.n), atol=1e-7)
            mask = gs.all(mask, axis=(1, 2))
            return mask

        raise ValueError('point_type should be \'vector\' or \'matrix\'.')

    @geomstats.vectorization.decorator(
        ['else', 'point', 'point_type', 'output_point'])
    def regularize(self, point, point_type=None):
        """Regularize a point to be in accordance with convention.

        In 3D, regularize the norm of the rotation vector,
        to be between 0 and pi, following the axis-angle
        representation's convention.

        If the angle angle is between pi and 2pi,
        the function computes its complementary in 2pi and
        inverts the direction of the rotation axis.

        Parameters
        ----------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        regularized_point : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        geomstats.error.check_parameter_accepted_values(
            point_type, 'point_type', ['vector', 'matrix'])
        if point_type == 'vector':
            regularized_point = point
            if self.n == 3:
                angle = gs.linalg.norm(regularized_point, axis=1)

                mask_0 = gs.isclose(angle, 0.)
                mask_not_0 = ~mask_0
                mask_pi = gs.isclose(angle, gs.pi)

                # This avoids division by 0.
                mask_0_float = gs.cast(mask_0, gs.float32) + self.epsilon
                mask_not_0_float = (
                    gs.cast(mask_not_0, gs.float32)
                    + self.epsilon)
                mask_pi_float = gs.cast(mask_pi, gs.float32) + self.epsilon

                k = gs.floor(angle / (2 * gs.pi) + .5)
                angle += mask_0_float

                norms_ratio = gs.zeros_like(angle)
                norms_ratio += mask_not_0_float * (
                    1. - 2. * gs.pi * k / angle)
                norms_ratio += mask_0_float
                norms_ratio += mask_pi_float * (
                    gs.pi / angle
                    - (1. - 2. * gs.pi * k / angle))

                regularized_point = gs.einsum(
                    '...,...i->...i', norms_ratio, regularized_point)

        elif point_type == 'matrix':
            regularized_point = point

        return regularized_point

    @geomstats.vectorization.decorator(
        ['else', 'point', 'else', 'point_type', 'output_point'])
    def regularize_tangent_vec_at_identity(
            self, tangent_vec, metric=None, point_type=None):
        """Regularize a tangent vector at the identify.

        In 3D, regularize a tangent_vector by getting its norm at the identity,
        determined by the metric, to be less than pi.

        Parameters
        ----------
        tangent_vec : array-like, shape=[n_samples, {dim, [n, n]}]
        metric : RiemannianMetric, optional
            default: self.left_canonical_metric
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        regularized_vec : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        geomstats.error.check_parameter_accepted_values(
            point_type, 'point_type', ['vector', 'matrix'])

        if point_type == 'vector':
            if self.n == 3:
                if metric is None:
                    metric = self.left_canonical_metric
                tangent_vec_metric_norm = metric.norm(tangent_vec)
                tangent_vec_canonical_norm = gs.linalg.norm(
                    tangent_vec, axis=-1)

                mask_norm_0 = gs.isclose(
                    tangent_vec_metric_norm, 0.)
                mask_canonical_norm_0 = gs.isclose(
                    tangent_vec_canonical_norm, 0.)

                mask_0 = mask_norm_0 | mask_canonical_norm_0
                mask_else = ~mask_0

                # This avoids division by 0.
                mask_0_float = gs.cast(mask_0, gs.float32) + self.epsilon
                mask_else_float = gs.cast(mask_else, gs.float32) + self.epsilon

                regularized_vec = gs.zeros_like(tangent_vec)
                regularized_vec += gs.einsum(
                    '...,...i->...i', mask_0_float, tangent_vec)

                tangent_vec_canonical_norm += mask_0_float

                coef = gs.zeros_like(tangent_vec_metric_norm)
                coef += mask_else_float * (
                    tangent_vec_metric_norm
                    / tangent_vec_canonical_norm)

                coef_tangent_vec = gs.einsum(
                    '...,...i->...i', coef, tangent_vec)
                regularized_vec += gs.einsum(
                    '...,...i->...i',
                    mask_else_float,
                    self.regularize(coef_tangent_vec))

                coef += mask_0_float
                regularized_vec = gs.einsum(
                    '...,...i->...i', 1. / coef, regularized_vec)
                regularized_vec = gs.einsum(
                    '...,...i->...i', mask_else_float, regularized_vec)
            else:
                # TODO(nina): Check if/how regularization is needed in nD?
                regularized_vec = tangent_vec

        elif point_type == 'matrix':
            is_vectorized = gs.ndim(gs.array(tangent_vec)) == 3
            axes = (0, 2, 1) if is_vectorized else (1, 0)
            regularized_vec = \
                1. / 2 * (tangent_vec - gs.transpose(tangent_vec, axes))

        return regularized_vec

    @geomstats.vectorization.decorator(
        ['else', 'point', 'point', 'else', 'point_type', 'output_point'])
    def regularize_tangent_vec(
            self, tangent_vec, base_point, metric=None, point_type=None):
        """Regularize tangent vector at a base point.

        In 3D, regularize a tangent_vector by getting the norm of its parallel
        transport to the identity, determined by the metric, less than pi.

        Parameters
        ----------
        tangent_vec : array-like, shape=[n_samples, {dim, [n, n]}]
            Tangent vector at base point.
        base_point : array-like, shape=[n_samples, {dim, [n, n]}]
            Point on the manifold.
        metric : RiemannianMetric, optional
            default: self.left_canonical_metric
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        regularized_tangent_vec : array-like,
            shape=[n_samples, {dim, [n, n]}]
        """
        geomstats.error.check_parameter_accepted_values(
            point_type, 'point_type', ['vector', 'matrix'])

        if point_type == 'vector':
            if self.n == 3:
                if metric is None:
                    metric = self.left_canonical_metric
                base_point = self.regularize(base_point, point_type)
                n_vecs = tangent_vec.shape[0]

                jacobian = self.jacobian_translation(
                    point=base_point,
                    left_or_right=metric.left_or_right,
                    point_type=point_type)
                jacobian = gs.array([jacobian[0]] * n_vecs)
                inv_jacobian = gs.linalg.inv(jacobian)
                inv_jacobian = gs.to_ndarray(inv_jacobian, to_ndim=3)
                tangent_vec_at_id = gs.einsum(
                    'ni,nij->nj',
                    tangent_vec,
                    gs.transpose(inv_jacobian, axes=(0, 2, 1)))

                tangent_vec_at_id = self.regularize_tangent_vec_at_identity(
                    tangent_vec_at_id, metric, point_type)

                jacobian = gs.to_ndarray(jacobian, to_ndim=3)
                regularized_tangent_vec = gs.einsum(
                    'ni,nij->nj',
                    tangent_vec_at_id,
                    gs.transpose(jacobian, axes=(0, 2, 1)))
            else:
                # TODO(nina): Check if/how regularization is needed in nD?
                regularized_tangent_vec = tangent_vec

        elif point_type == 'matrix':
            if gs.allclose(base_point, self.identity):
                return self.regularize_tangent_vec_at_identity(
                    tangent_vec, point_type=point_type)

            inv_base_point = self.inverse(base_point)
            tangent_vec_at_id = self.compose(inv_base_point, tangent_vec)
            regularized_tangent_vec = self.regularize_tangent_vec_at_identity(
                tangent_vec_at_id, point_type=point_type)
            regularized_tangent_vec = self.compose(
                base_point, regularized_tangent_vec)

        return regularized_tangent_vec

    @geomstats.vectorization.decorator(['else', 'matrix'])
    def projection(self, point):
        """Project a matrix on SO(n) using the Frobenius norm.

        Parameters
        ----------
        mat : array-like, shape=[n_samples, n, n]

        Returns
        -------
        rot_mat : array-like, shape=[n_samples, n, n]
        """
        mat = point
        n_mats, _, _ = mat.shape

        if self.n == 3:
            mat_unitary_u, _, mat_unitary_v = gs.linalg.svd(mat)
            rot_mat = gs.einsum('nij,njk->nik', mat_unitary_u, mat_unitary_v)
            mask = gs.less(gs.linalg.det(rot_mat), 0.)
            mask_float = gs.cast(mask, gs.float32) + self.epsilon
            diag = gs.array([[1., 1., -1.]])
            diag = gs.to_ndarray(
                algebra_utils.from_vector_to_diagonal_matrix(diag),
                to_ndim=3) + self.epsilon
            new_mat_diag_s = gs.tile(diag, [n_mats, 1, 1])

            aux_mat = gs.einsum(
                'nij,njk->nik',
                mat_unitary_u,
                new_mat_diag_s)
            rot_mat += gs.einsum(
                'n,njk->njk',
                mask_float,
                gs.einsum(
                    'nij,njk->nik',
                    aux_mat,
                    mat_unitary_v))
        else:
            aux_mat = gs.matmul(gs.transpose(mat, axes=(0, 2, 1)), mat)

            inv_sqrt_mat = gs.linalg.inv(
                gs.linalg.sqrtm(aux_mat))

            rot_mat = gs.matmul(mat, inv_sqrt_mat)

        return rot_mat

    @geomstats.vectorization.decorator(['else', 'vector'])
    def skew_matrix_from_vector(self, vec):
        """Get the skew-symmetric matrix derived from the vector.

        In 3D, compute the skew-symmetric matrix,known as the cross-product of
        a vector, associated to the vector `vec`.

        In nD, fill a skew-symmetric matrix with the values of the vector.

        Parameters
        ----------
        vec : array-like, shape=[n_samples, dim]

        Returns
        -------
        skew_mat : array-like, shape=[n_samples, n, n]
        """
        n_vecs, vec_dim = gs.shape(vec)

        if self.n == 2:  # SO(2)
            vec = gs.tile(vec, [1, 2])
            vec = gs.reshape(vec, (n_vecs, 2))

            id_skew = gs.array(
                gs.tile([[[0., 1.], [-1., 0.]]], (n_vecs, 1, 1)))
            skew_mat = gs.einsum(
                '...ij,...i->...ij', gs.cast(id_skew, gs.float32), vec)

        elif self.n == 3:  # SO(3)
            # This avoids dividing by 0.
            levi_civita_symbol = gs.tile([[
                [[0., 0., 0.],
                 [0., 0., 1.],
                 [0., -1., 0.]],
                [[0., 0., -1.],
                 [0., 0., 0.],
                 [1., 0., 0.]],
                [[0., 1., 0.],
                 [-1., 0., 0.],
                 [0., 0., 0.]]
            ]], (n_vecs, 1, 1, 1))

            levi_civita_symbol = gs.array(levi_civita_symbol)
            levi_civita_symbol += self.epsilon

            # This avoids dividing by 0.
            basis_vec_1 = gs.array(
                gs.tile([[1., 0., 0.]], (n_vecs, 1))) + self.epsilon
            basis_vec_2 = gs.array(
                gs.tile([[0., 1., 0.]], (n_vecs, 1))) + self.epsilon
            basis_vec_3 = gs.array(
                gs.tile([[0., 0., 1.]], (n_vecs, 1))) + self.epsilon

            cross_prod_1 = gs.einsum(
                'nijk,ni,nj->nk',
                levi_civita_symbol,
                basis_vec_1,
                vec)
            cross_prod_2 = gs.einsum(
                'nijk,ni,nj->nk',
                levi_civita_symbol,
                basis_vec_2,
                vec)
            cross_prod_3 = gs.einsum(
                'nijk,ni,nj->nk',
                levi_civita_symbol,
                basis_vec_3,
                vec)

            cross_prod_1 = gs.to_ndarray(cross_prod_1, to_ndim=3, axis=1)
            cross_prod_2 = gs.to_ndarray(cross_prod_2, to_ndim=3, axis=1)
            cross_prod_3 = gs.to_ndarray(cross_prod_3, to_ndim=3, axis=1)
            skew_mat = gs.concatenate(
                [cross_prod_1, cross_prod_2, cross_prod_3], axis=1)

        else:  # SO(n)
            mat_dim = gs.cast(
                ((1. + gs.sqrt(1. + 8. * vec_dim)) / 2.), gs.int32)
            skew_mat = gs.zeros((n_vecs,) + (self.n,) * 2)
            upper_triangle_indices = gs.triu_indices(mat_dim, k=1)
            for i in range(n_vecs):
                skew_mat[i][upper_triangle_indices] = vec[i]
                skew_mat[i] = skew_mat[i] - gs.transpose(skew_mat[i])
        return skew_mat

    @geomstats.vectorization.decorator(['else', 'matrix', 'output_point'])
    def vector_from_skew_matrix(self, skew_mat):
        """Derive a vector from the skew-symmetric matrix.

        In 3D, compute the vector defining the cross product
        associated to the skew-symmetric matrix skew mat.

        In nD, fill a vector by reading the values
        of the upper triangle of skew_mat.

        Parameters
        ----------
        skew_mat : array-like, shape=[n_samples, n, n]

        Returns
        -------
        vec : array-like, shape=[n_samples, dim]
        """
        n_skew_mats, mat_dim_1, _ = skew_mat.shape

        vec_dim = self.dim
        vec = gs.zeros((n_skew_mats, vec_dim))

        if self.n == 2:  # SO(2)
            vec = skew_mat[:, 0, 1]
            vec = gs.expand_dims(vec, axis=1)

        elif self.n == 3:  # SO(3)
            vec_1 = gs.to_ndarray(skew_mat[:, 2, 1], to_ndim=2, axis=1)
            vec_2 = gs.to_ndarray(skew_mat[:, 0, 2], to_ndim=2, axis=1)
            vec_3 = gs.to_ndarray(skew_mat[:, 1, 0], to_ndim=2, axis=1)
            vec = gs.concatenate([vec_1, vec_2, vec_3], axis=1)

        else:  # SO(n)
            idx = 0
            for j in range(mat_dim_1):
                for i in range(j):
                    vec[:, idx] = skew_mat[:, i, j]
                    idx += 1

        return vec

    @geomstats.vectorization.decorator(['else', 'matrix', 'output_point'])
    def rotation_vector_from_matrix(self, rot_mat):
        r"""Convert rotation matrix (in 3D) to rotation vector (axis-angle).

        Get the angle through the trace of the rotation matrix:
        The eigenvalues are:
        :math:`\{1, \cos(angle) + i \sin(angle), \cos(angle) - i \sin(angle)\}`
        so that:
        :math:`trace = 1 + 2 \cos(angle), \{-1 \leq trace \leq 3\}`

        Get the rotation vector through the formula:
        :math:`S_r = \frac{angle}{(2 * \sin(angle) ) (R - R^T)}`

        For the edge case where the angle is close to pi,
        the formulation is derived by using the following equality (see the
        Axis-angle representation on Wikipedia):
        :math:`outer(r, r) = \frac{1}{2} (R + I_3)`

        In nD, the rotation vector stores the :math:`n(n-1)/2` values
        of the skew-symmetric matrix representing the rotation.

        Parameters
        ----------
        rot_mat : array-like, shape=[n_samples, n, n]

        Returns
        -------
        regularized_rot_vec : array-like, shape=[n_samples, dim]
        """
        n_rot_mats, _, _ = rot_mat.shape

        if self.n == 3:
            trace = gs.trace(rot_mat, axis1=1, axis2=2)
            trace = gs.to_ndarray(trace, to_ndim=2, axis=1)
            trace_num = gs.clip(trace, -1, 3)
            angle = gs.arccos(0.5 * (trace_num - 1))
            rot_mat_transpose = gs.transpose(rot_mat, axes=(0, 2, 1))
            rot_vec_not_pi = self.vector_from_skew_matrix(
                rot_mat - rot_mat_transpose)
            mask_0 = gs.cast(gs.isclose(angle, 0.), gs.float32)
            mask_pi = gs.cast(gs.isclose(angle, gs.pi, atol=1e-2), gs.float32)
            mask_else = (1 - mask_0) * (1 - mask_pi)

            numerator = 0.5 * mask_0 + angle * mask_else
            denominator = (1 - angle ** 2 / 6) * mask_0 + 2 * gs.sin(
                angle) * mask_else + mask_pi

            rot_vec_not_pi = rot_vec_not_pi * numerator / denominator

            vector_outer = 0.5 * (gs.eye(3) + rot_mat)
            gs.set_diag(
                vector_outer, gs.maximum(
                    0., gs.diagonal(vector_outer, axis1=1, axis2=2)))
            squared_diag_comp = gs.diagonal(vector_outer, axis1=1, axis2=2)
            diag_comp = gs.sqrt(squared_diag_comp)
            norm_line = gs.linalg.norm(vector_outer, axis=2)
            max_line_index = gs.argmax(norm_line, axis=1)
            selected_line = gs.get_slice(
                vector_outer, (range(n_rot_mats), max_line_index))
            signs = gs.sign(selected_line)
            rot_vec_pi = angle * signs * diag_comp

            rot_vec = rot_vec_not_pi + mask_pi * rot_vec_pi

        else:
            skew_mat = self.embedding_manifold.log(rot_mat)
            rot_vec = self.vector_from_skew_matrix(skew_mat)

        return self.regularize(rot_vec, point_type='vector')

    @geomstats.vectorization.decorator(['else', 'vector'])
    def matrix_from_rotation_vector(self, rot_vec):
        """Convert rotation vector to rotation matrix.

        Parameters
        ----------
        rot_vec: array-like, shape=[n_samples, dim]

        Returns
        -------
        rot_mat: array-like, shape=[n_samples, {dim, [n, n]}]
        """
        rot_vec = self.regularize(rot_vec, point_type='vector')

        if self.n == 3:
            angle = gs.linalg.norm(rot_vec, axis=1)
            angle = gs.to_ndarray(angle, to_ndim=2, axis=1)

            skew_rot_vec = self.skew_matrix_from_vector(rot_vec)

            coef_1 = gs.zeros_like(angle)
            coef_2 = gs.zeros_like(angle)

            # This avoids dividing by 0.
            mask_0 = gs.isclose(angle, 0.)
            mask_0_float = gs.cast(mask_0, gs.float32) + self.epsilon

            coef_1 += mask_0_float * (1. - (angle ** 2) / 6.)
            coef_2 += mask_0_float * (1. / 2. - angle ** 2)

            # This avoids dividing by 0.
            mask_else = ~mask_0
            mask_else_float = gs.cast(mask_else, gs.float32) + self.epsilon

            angle += mask_0_float

            coef_1 += mask_else_float * (gs.sin(angle) / angle)
            coef_2 += mask_else_float * (
                (1. - gs.cos(angle)) / (angle ** 2))

            coef_1 = gs.squeeze(coef_1, axis=1)
            coef_2 = gs.squeeze(coef_2, axis=1)
            term_1 = (gs.eye(self.dim)
                      + gs.einsum('n,njk->njk', coef_1, skew_rot_vec))

            squared_skew_rot_vec = gs.einsum(
                'nij,njk->nik', skew_rot_vec, skew_rot_vec)

            term_2 = gs.einsum('n,njk->njk', coef_2, squared_skew_rot_vec)

            rot_mat = term_1 + term_2

        else:
            skew_mat = self.skew_matrix_from_vector(rot_vec)
            rot_mat = self.embedding_manifold.exp(skew_mat)

        return rot_mat

    @geomstats.vectorization.decorator(['else', 'matrix'])
    def quaternion_from_matrix(self, rot_mat):
        """Convert a rotation matrix into a unit quaternion.

        Parameters
        ----------
        rot_mat : array-like, shape=[n_samples, n, n]

        Returns
        -------
        quaternion : array-like, shape=[n_samples, 4]
        """
        if self.n != 3:
            raise ValueError(
                'The quaternion representation does not exist'
                ' for rotations in %d dimensions.' % self.n)
        rot_vec = self.rotation_vector_from_matrix(rot_mat)
        quaternion = self.quaternion_from_rotation_vector(rot_vec)

        return quaternion

    @geomstats.vectorization.decorator(['else', 'vector'])
    def quaternion_from_rotation_vector(self, rot_vec):
        """Convert a rotation vector into a unit quaternion.

        Parameters
        ----------
        rot_vec : array-like, shape=[n_samples, dim]

        Returns
        -------
        quaternion : array-like, shape=[n_samples, 4]
        """
        if self.n != 3:
            raise ValueError(
                'The quaternion representation does not exist'
                ' for rotations in %d dimensions.' % self.n)
        rot_vec = self.regularize(rot_vec, point_type='vector')

        angle = gs.linalg.norm(rot_vec, axis=1)
        angle = gs.to_ndarray(angle, to_ndim=2, axis=1)

        mask_0 = gs.isclose(angle, 0.)
        mask_not_0 = ~mask_0

        rotation_axis = gs.divide(
            rot_vec,
            angle
            * gs.cast(mask_not_0, gs.float32)
            + gs.cast(mask_0, gs.float32))

        quaternion = gs.concatenate(
            (gs.cos(angle / 2),
             gs.sin(angle / 2) * rotation_axis[:]),
            axis=1)

        return quaternion

    @geomstats.vectorization.decorator(['else', 'vector'])
    def rotation_vector_from_quaternion(self, quaternion):
        """Convert a unit quaternion into a rotation vector.

        Parameters
        ----------
        quaternion : array-like, shape=[n_samples, 4]

        Returns
        -------
        rot_vec : array-like, shape=[n_samples, dim]
        """
        if self.n != 3:
            raise ValueError(
                'The quaternion representation does not exist'
                ' for rotations in %d dimensions.' % self.n)

        cos_half_angle = quaternion[:, 0]
        cos_half_angle = gs.clip(cos_half_angle, -1, 1)
        half_angle = gs.arccos(cos_half_angle)

        half_angle = gs.to_ndarray(half_angle, to_ndim=2, axis=1)

        mask_0 = gs.isclose(half_angle, 0.)
        mask_not_0 = ~mask_0

        rotation_axis = gs.divide(
            quaternion[:, 1:],
            gs.sin(half_angle) *
            gs.cast(mask_not_0, gs.float32)
            + gs.cast(mask_0, gs.float32))
        rot_vec = gs.array(
            2 * half_angle
            * rotation_axis
            * gs.cast(mask_not_0, gs.float32))

        rot_vec = self.regularize(rot_vec, point_type='vector')
        return rot_vec

    @geomstats.vectorization.decorator(['else', 'vector'])
    def matrix_from_quaternion(self, quaternion):
        """Convert a unit quaternion into a rotation vector.

        Parameters
        ----------
        quaternion : array-like, shape=[n_samples, 4]

        Returns
        -------
        rot_mat : array-like, shape=[n_samples, dim]
        """
        if self.n != 3:
            raise ValueError(
                'The quaternion representation does not exist'
                ' for rotations in %d dimensions.' % self.n)
        n_quaternions, _ = quaternion.shape

        w, x, y, z = gs.hsplit(quaternion, 4)

        rot_mat = gs.zeros((n_quaternions,) + (self.n,) * 2)

        for i in range(n_quaternions):
            # TODO(nina): Vectorize by applying the composition of
            # quaternions to the identity matrix
            column_1 = [w[i] ** 2 + x[i] ** 2 - y[i] ** 2 - z[i] ** 2,
                        2 * x[i] * y[i] - 2 * w[i] * z[i],
                        2 * x[i] * z[i] + 2 * w[i] * y[i]]

            column_2 = [2 * x[i] * y[i] + 2 * w[i] * z[i],
                        w[i] ** 2 - x[i] ** 2 + y[i] ** 2 - z[i] ** 2,
                        2 * y[i] * z[i] - 2 * w[i] * x[i]]

            column_3 = [2 * x[i] * z[i] - 2 * w[i] * y[i],
                        2 * y[i] * z[i] + 2 * w[i] * x[i],
                        w[i] ** 2 - x[i] ** 2 - y[i] ** 2 + z[i] ** 2]

            mask_i = gs.get_mask_i_float(i, n_quaternions)
            rot_mat_i = gs.transpose(
                gs.hstack([column_1, column_2, column_3]))
            rot_mat_i = gs.to_ndarray(rot_mat_i, to_ndim=3)
            rot_mat += gs.einsum('n,nij->nij', mask_i, rot_mat_i)

        return rot_mat

    @geomstats.vectorization.decorator(['else', 'vector'])
    def matrix_from_tait_bryan_angles_extrinsic_xyz(self, tait_bryan_angles):
        """Convert Tait-Bryan angles to rot mat in extrensic coords (xyz).

        Convert a rotation given in terms of the tait bryan angles,
        [angle_1, angle_2, angle_3] in extrinsic (fixed) coordinate system
        in order xyz, into a rotation matrix.

        rot_mat = Z(angle_1).Y(angle_2).X(angle_3)
        where:
        - Z(angle_1) is a rotation of angle angle_1 around axis z.
        - Y(angle_2) is a rotation of angle angle_2 around axis y.
        - X(angle_3) is a rotation of angle angle_3 around axis x.

        Parameters
        ----------
        tait_bryan_angles : array-like, shape=[n_samples, 3]

        Returns
        -------
        rot_mat : array-like, shape=[n_samples, n, n]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' does not exist for rotations in %d dimensions.' % self.n)
        n_tait_bryan_angles, _ = tait_bryan_angles.shape

        rot_mat = gs.zeros((n_tait_bryan_angles,) + (self.n,) * 2)
        angle_1 = tait_bryan_angles[:, 0]
        angle_2 = tait_bryan_angles[:, 1]
        angle_3 = tait_bryan_angles[:, 2]

        for i in range(n_tait_bryan_angles):
            cos_angle_1 = gs.cos(angle_1[i])
            sin_angle_1 = gs.sin(angle_1[i])
            cos_angle_2 = gs.cos(angle_2[i])
            sin_angle_2 = gs.sin(angle_2[i])
            cos_angle_3 = gs.cos(angle_3[i])
            sin_angle_3 = gs.sin(angle_3[i])

            column_1 = [[cos_angle_1 * cos_angle_2],
                        [cos_angle_2 * sin_angle_1],
                        [- sin_angle_2]]
            column_2 = [[(cos_angle_1 * sin_angle_2 * sin_angle_3
                          - cos_angle_3 * sin_angle_1)],
                        [(cos_angle_1 * cos_angle_3
                          + sin_angle_1 * sin_angle_2 * sin_angle_3)],
                        [cos_angle_2 * sin_angle_3]]
            column_3 = [[(sin_angle_1 * sin_angle_3
                          + cos_angle_1 * cos_angle_3 * sin_angle_2)],
                        [(cos_angle_3 * sin_angle_1 * sin_angle_2
                          - cos_angle_1 * sin_angle_3)],
                        [cos_angle_2 * cos_angle_3]]

            rot_mat[i] = gs.hstack((column_1, column_2, column_3))
        return rot_mat

    @geomstats.vectorization.decorator(['else', 'vector'])
    def matrix_from_tait_bryan_angles_extrinsic_zyx(self, tait_bryan_angles):
        """Convert Tait-Bryan angles to rot mat in extrensic coords (zyx).

        Convert a rotation given in terms of the tait bryan angles,
        [angle_1, angle_2, angle_3] in extrinsic (fixed) coordinate system
        in order zyx, into a rotation matrix.

        rot_mat = X(angle_1).Y(angle_2).Z(angle_3)
        where:
        - X(angle_1) is a rotation of angle angle_1 around axis x.
        - Y(angle_2) is a rotation of angle angle_2 around axis y.
        - Z(angle_3) is a rotation of angle angle_3 around axis z.

        Parameters
        ----------
        tait_bryan_angles : array-like, shape=[n_samples, 3]

        Returns
        -------
        rot_mat : array-like, shape=[n_samples, n, n]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' does not exist for rotations in %d dimensions.' % self.n)
        n_tait_bryan_angles, _ = tait_bryan_angles.shape

        rot_mat = gs.zeros((n_tait_bryan_angles,) + (self.n,) * 2)
        angle_1 = tait_bryan_angles[:, 0]
        angle_2 = tait_bryan_angles[:, 1]
        angle_3 = tait_bryan_angles[:, 2]

        for i in range(n_tait_bryan_angles):
            cos_angle_1 = gs.cos(angle_1[i])
            sin_angle_1 = gs.sin(angle_1[i])
            cos_angle_2 = gs.cos(angle_2[i])
            sin_angle_2 = gs.sin(angle_2[i])
            cos_angle_3 = gs.cos(angle_3[i])
            sin_angle_3 = gs.sin(angle_3[i])

            column_1 = [[cos_angle_2 * cos_angle_3],
                        [(cos_angle_1 * sin_angle_3
                          + cos_angle_3 * sin_angle_1 * sin_angle_2)],
                        [(sin_angle_1 * sin_angle_3
                          - cos_angle_1 * cos_angle_3 * sin_angle_2)]]

            column_2 = [[- cos_angle_2 * sin_angle_3],
                        [(cos_angle_1 * cos_angle_3
                          - sin_angle_1 * sin_angle_2 * sin_angle_3)],
                        [(cos_angle_3 * sin_angle_1
                          + cos_angle_1 * sin_angle_2 * sin_angle_3)]]

            column_3 = [[sin_angle_2],
                        [- cos_angle_2 * sin_angle_1],
                        [cos_angle_1 * cos_angle_2]]
            rot_mat[i] = gs.hstack((column_1, column_2, column_3))
        return rot_mat

    @geomstats.vectorization.decorator(['else', 'vector', 'else', 'else'])
    def matrix_from_tait_bryan_angles(self, tait_bryan_angles,
                                      extrinsic_or_intrinsic='extrinsic',
                                      order='zyx'):
        """Convert Tait-Bryan angles to rot mat in extr or intr coords.

        Convert a rotation given in terms of the tait bryan angles,
        [angle_1, angle_2, angle_3] in extrinsic (fixed) or
        intrinsic (moving) coordinate frame into a rotation matrix.

        If the order is zyx, into the rotation matrix rot_mat:
        rot_mat = X(angle_1).Y(angle_2).Z(angle_3)
        where:
        - X(angle_1) is a rotation of angle angle_1 around axis x.
        - Y(angle_2) is a rotation of angle angle_2 around axis y.
        - Z(angle_3) is a rotation of angle angle_3 around axis z.

        Exchanging 'extrinsic' and 'intrinsic' amounts to
        exchanging the order.

        Parameters
        ----------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        extrinsic_or_intrinsic : str, {'extrensic', 'intrinsic'} optional
            default: 'extrinsic'
        order : str, {'xyz', 'zyx'}, optional
            default: 'zyx'

        Returns
        -------
        rot_mat : array-like, shape=[n_samples, n, n]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' does not exist for rotations in %d dimensions.' % self.n)

        extrinsic_zyx = (extrinsic_or_intrinsic == 'extrinsic'
                         and order == 'zyx')
        intrinsic_xyz = (extrinsic_or_intrinsic == 'intrinsic'
                         and order == 'xyz')

        extrinsic_xyz = (extrinsic_or_intrinsic == 'extrinsic'
                         and order == 'xyz')
        intrinsic_zyx = (extrinsic_or_intrinsic == 'intrinsic'
                         and order == 'zyx')

        if extrinsic_zyx:
            rot_mat = self.matrix_from_tait_bryan_angles_extrinsic_zyx(
                tait_bryan_angles)
        elif intrinsic_xyz:
            tait_bryan_angles_reversed = gs.flip(tait_bryan_angles, axis=1)
            rot_mat = self.matrix_from_tait_bryan_angles_extrinsic_zyx(
                tait_bryan_angles_reversed)

        elif extrinsic_xyz:
            rot_mat = self.matrix_from_tait_bryan_angles_extrinsic_xyz(
                tait_bryan_angles)
        elif intrinsic_zyx:
            tait_bryan_angles_reversed = gs.flip(tait_bryan_angles, axis=1)
            rot_mat = self.matrix_from_tait_bryan_angles_extrinsic_xyz(
                tait_bryan_angles_reversed)

        else:
            raise ValueError('extrinsic_or_intrinsic should be'
                             ' \'extrinsic\' or \'intrinsic\''
                             ' and order should be \'xyz\' or \'zyx\'.')

        return rot_mat

    @geomstats.vectorization.decorator(['else', 'matrix', 'else', 'else'])
    def tait_bryan_angles_from_matrix(self, rot_mat,
                                      extrinsic_or_intrinsic='extrinsic',
                                      order='zyx'):
        """Convert rot_mat into Tait-Bryan angles.

        Convert a rotation matrix rot_mat into the tait bryan angles,
        [angle_1, angle_2, angle_3] in extrinsic (fixed) coordinate frame,
        for the order zyx, i.e.:
        rot_mat = X(angle_1).Y(angle_2).Z(angle_3)
        where:
        - X(angle_1) is a rotation of angle angle_1 around axis x.
        - Y(angle_2) is a rotation of angle angle_2 around axis y.
        - Z(angle_3) is a rotation of angle angle_3 around axis z.

        Parameters
        ----------
        rot_mat : array-like, shape=[n_samples, n, n]
        extrinsic_or_intrinsic : str, {'extrinsic', 'intrinsic'}, optional
            default: 'extrinsic'
        order : str, {'xyz', 'zyx'}, optional
            default: 'zyx'

        Returns
        -------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' does not exist for rotations in %d dimensions.' % self.n)

        quaternion = self.quaternion_from_matrix(rot_mat)
        tait_bryan_angles = self.tait_bryan_angles_from_quaternion(
            quaternion,
            extrinsic_or_intrinsic=extrinsic_or_intrinsic,
            order=order)

        return tait_bryan_angles

    @geomstats.vectorization.decorator(['else', 'vector'])
    def quaternion_from_tait_bryan_angles_intrinsic_xyz(
            self, tait_bryan_angles):
        """Convert Tait-Bryan angles to into unit quaternion.

        Convert a rotation given by Tait-Bryan angles in extrinsic
        coordinate systems and order xyz into a unit quaternion.

        Parameters
        ----------
        tait_bryan_angles : array-like, shape=[n_samples, 3]

        Returns
        -------
        quaternion : array-like, shape=[n_samples, 4]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' and the quaternion representation'
                ' do not exist for rotations in %d dimensions.' % self.n)
        n_tait_bryan_angles, _ = tait_bryan_angles.shape
        quaternion = gs.zeros((n_tait_bryan_angles, 4))

        matrix = self.matrix_from_tait_bryan_angles(
            tait_bryan_angles,
            extrinsic_or_intrinsic='intrinsic',
            order='xyz')
        quaternion = self.quaternion_from_matrix(matrix)
        return quaternion

    @geomstats.vectorization.decorator(['else', 'vector', 'else', 'else'])
    def quaternion_from_tait_bryan_angles(self, tait_bryan_angles,
                                          extrinsic_or_intrinsic='extrinsic',
                                          order='zyx'):
        """Convert a rotation given by Tait-Bryan angles into unit quaternion.

        Parameters
        ----------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        extrinsic_or_intrinsic : str, {'extrinsic', 'intrinsic'}, optional
            default: 'extrinsic'
        order : str, {'xyz', 'zyx'}, optional
            default: 'zyx'

        Returns
        -------
        quat : array-like, shape=[n_samples, 4]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' and the quaternion representation'
                ' do not exist for rotations in %d dimensions.' % self.n)

        extrinsic_zyx = (extrinsic_or_intrinsic == 'extrinsic'
                         and order == 'zyx')
        intrinsic_xyz = (extrinsic_or_intrinsic == 'intrinsic'
                         and order == 'xyz')

        extrinsic_xyz = (extrinsic_or_intrinsic == 'extrinsic'
                         and order == 'xyz')
        intrinsic_zyx = (extrinsic_or_intrinsic == 'intrinsic'
                         and order == 'zyx')

        if extrinsic_zyx:
            tait_bryan_angles_reversed = gs.flip(tait_bryan_angles, axis=1)
            quat = self.quaternion_from_tait_bryan_angles_intrinsic_xyz(
                tait_bryan_angles_reversed)

        elif intrinsic_xyz:
            quat = self.quaternion_from_tait_bryan_angles_intrinsic_xyz(
                tait_bryan_angles)

        elif extrinsic_xyz:
            rot_mat = self.matrix_from_tait_bryan_angles_extrinsic_xyz(
                tait_bryan_angles)
            quat = self.quaternion_from_matrix(rot_mat)

        elif intrinsic_zyx:
            tait_bryan_angles_reversed = gs.flip(tait_bryan_angles, axis=1)
            rot_mat = self.matrix_from_tait_bryan_angles_extrinsic_xyz(
                tait_bryan_angles_reversed)
            quat = self.quaternion_from_matrix(rot_mat)
        else:
            raise ValueError('extrinsic_or_intrinsic should be'
                             ' \'extrinsic\' or \'intrinsic\''
                             ' and order should be \'xyz\' or \'zyx\'.')

        return quat

    @geomstats.vectorization.decorator(['else', 'vector', 'else', 'else'])
    def rotation_vector_from_tait_bryan_angles(
            self,
            tait_bryan_angles,
            extrinsic_or_intrinsic='extrinsic',
            order='zyx'):
        """Convert rotation given by angle_1, angle_2, angle_3 into rot. vec.

        Convert into axis-angle representation.

        Parameters
        ----------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        extrinsic_or_intrinsic : str, {'extrinsic', 'intrinsic'}, optional
            default: 'extrinsic'
        order : str, {'xyz', 'zyx'}, optional
            default: 'zyx'

        Returns
        -------
        rot_vec : array-like, shape=[n_samples, dim]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' does not exist for rotations in %d dimensions.' % self.n)

        quaternion = self.quaternion_from_tait_bryan_angles(
            tait_bryan_angles,
            extrinsic_or_intrinsic=extrinsic_or_intrinsic,
            order=order)
        rot_vec = self.rotation_vector_from_quaternion(quaternion)

        rot_vec = self.regularize(rot_vec, point_type='vector')
        return rot_vec

    @geomstats.vectorization.decorator(['else', 'vector'])
    def tait_bryan_angles_from_quaternion_intrinsic_zyx(self, quaternion):
        """Convert quaternion to tait bryan representation of order zyx.

        Parameters
        ----------
        quaternion : array-like, shape=[n_samples, 4]

        Returns
        -------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' and the quaternion representation'
                ' do not exist for rotations in %d dimensions.' % self.n)

        w, x, y, z = gs.hsplit(quaternion, 4)
        angle_1 = gs.arctan2(y * z + w * x,
                             1. / 2. - (x ** 2 + y ** 2))
        angle_2 = gs.arcsin(- 2. * (x * z - w * y))
        angle_3 = gs.arctan2(x * y + w * z,
                             1. / 2. - (y ** 2 + z ** 2))
        tait_bryan_angles = gs.concatenate(
            [angle_1, angle_2, angle_3], axis=1)
        return tait_bryan_angles

    @geomstats.vectorization.decorator(['else', 'vector'])
    def tait_bryan_angles_from_quaternion_intrinsic_xyz(self, quaternion):
        """Convert quaternion to tait bryan representation of order xyz.

        Parameters
        ----------
        quaternion : array-like, shape=[n_samples, 4]

        Returns
        -------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' and the quaternion representation'
                ' do not exist for rotations in %d dimensions.' % self.n)

        w, x, y, z = gs.hsplit(quaternion, 4)

        angle_1 = gs.arctan2(2. * (- x * y + w * z),
                             w * w + x * x - y * y - z * z)
        angle_2 = gs.arcsin(2 * (x * z + w * y))
        angle_3 = gs.arctan2(2. * (- y * z + w * x),
                             w * w + z * z - x * x - y * y)

        tait_bryan_angles = gs.concatenate(
            [angle_1, angle_2, angle_3], axis=1)
        return tait_bryan_angles

    @geomstats.vectorization.decorator(['else', 'vector', 'else', 'else'])
    def tait_bryan_angles_from_quaternion(
            self, quaternion, extrinsic_or_intrinsic='extrinsic', order='zyx'):
        """Convert quaternion to a rotation in form angle_1, angle_2, angle_3.

        Parameters
        ----------
        quaternion : array-like, shape=[n_samples, 4]
        extrinsic_or_intrinsic : str, {'extrinsic', 'intrinsic'}, optional
            default: 'extrinsic'
        order : str, {'xyz', 'zyx'}, optional
            default: 'zyx'

        Returns
        -------
        tait_bryan : array-like, shape=[n_samples, 3]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' and the quaternion representation'
                ' do not exist for rotations in %d dimensions.' % self.n)

        extrinsic_zyx = (extrinsic_or_intrinsic == 'extrinsic'
                         and order == 'zyx')
        intrinsic_xyz = (extrinsic_or_intrinsic == 'intrinsic'
                         and order == 'xyz')

        extrinsic_xyz = (extrinsic_or_intrinsic == 'extrinsic'
                         and order == 'xyz')
        intrinsic_zyx = (extrinsic_or_intrinsic == 'intrinsic'
                         and order == 'zyx')

        if extrinsic_zyx:
            tait_bryan = self.tait_bryan_angles_from_quaternion_intrinsic_xyz(
                quaternion)
            tait_bryan = gs.flip(tait_bryan, axis=1)
        elif intrinsic_xyz:
            tait_bryan = self.tait_bryan_angles_from_quaternion_intrinsic_xyz(
                quaternion)

        elif extrinsic_xyz:
            tait_bryan = self.tait_bryan_angles_from_quaternion_intrinsic_zyx(
                quaternion)
            tait_bryan = gs.flip(tait_bryan, axis=1)
        elif intrinsic_zyx:
            tait_bryan = self.tait_bryan_angles_from_quaternion_intrinsic_zyx(
                quaternion)

        else:
            raise ValueError('extrinsic_or_intrinsic should be'
                             ' \'extrinsic\' or \'intrinsic\''
                             ' and order should be \'xyz\' or \'zyx\'.')

        return tait_bryan

    @geomstats.vectorization.decorator(['else', 'vector', 'else', 'else'])
    def tait_bryan_angles_from_rotation_vector(
            self, rot_vec, extrinsic_or_intrinsic='extrinsic', order='zyx'):
        """Convert a rotation vector to a rotation given by Tait-Bryan angles.

        Here the rotation vector is in the axis-angle representation.

        Parameters
        ----------
        rot_vec : array-like, shape=[n_samples, dim]
        extrinsic_or_intrinsic : str, {'extrinsic', 'intrinsic'}, optional
            default: 'extrinsic'
        order : str, {'xyz', 'zyx'}, optional
            default: 'zyx'

        Returns
        -------
        tait_bryan_angles : array-like, shape=[n_samples, 3]
        """
        if self.n != 3:
            raise ValueError(
                'The Tait-Bryan angles representation'
                ' does not exist for rotations in %d dimensions.' % self.n)

        quaternion = self.quaternion_from_rotation_vector(rot_vec)
        tait_bryan_angles = self.tait_bryan_angles_from_quaternion(
            quaternion,
            extrinsic_or_intrinsic=extrinsic_or_intrinsic,
            order=order)

        return tait_bryan_angles

    @geomstats.vectorization.decorator(
        ['else', 'point', 'point', 'point_type', 'output_point'])
    def compose(self, point_a, point_b, point_type=None):
        """Compose two elements of SO(n).

        Parameters
        ----------
        point_1 : array-like, shape=[n_samples, {dim, [n, n]}]
        point_2 : array-like, shape=[n_samples, {dim, [n, n]}]
        point_type : str, {'vector', 'matrix'}, optional
            default: default_point_type

        Returns
        -------
        point_prod : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        point_a = self.regularize(point_a, point_type=point_type)
        point_b = self.regularize(point_b, point_type=point_type)

        if point_type == 'vector':
            point_a = self.matrix_from_rotation_vector(point_a)
            point_b = self.matrix_from_rotation_vector(point_b)

        point_prod = gs.einsum('...jk,...kl->...jl', point_a, point_b)

        if point_type == 'vector':
            point_prod = self.rotation_vector_from_matrix(point_prod)

        point_prod = self.regularize(
            point_prod, point_type=point_type)
        return point_prod

    def inverse(self, point, point_type=None):
        """Compute the group inverse in SO(n).

        Parameters
        ----------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        inv_point : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        if point_type is None:
            point_type = self.default_point_type

        if point_type == 'vector':
            if self.n == 3:
                return -self.regularize(point, point_type=point_type)
            point = self.matrix_from_rotation_vector(point)

        transpose_order = (0, 2, 1) if gs.ndim(point) == 3 else (1, 0)
        inv_point = gs.transpose(point, transpose_order)

        if point_type == 'vector':
            inv_point = self.rotation_vector_from_matrix(inv_point)

        return inv_point

    @geomstats.vectorization.decorator(['else', 'point', 'else', 'point_type'])
    def jacobian_translation(
            self, point, left_or_right='left', point_type=None):
        """Compute the jacobian matrix corresponding to translation.

        Compute the jacobian matrix of the differential
        of the left/right translations from the identity to point in SO(n).

        Parameters
        ----------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
        left_or_right : str, {'left', 'right'}, optional
            default: 'left'
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        jacobian : array-like, shape=[n_samples, dim, dim]
        """
        geomstats.error.check_parameter_accepted_values(
            left_or_right, 'left_or_right', ['left', 'right'])

        if point_type == 'vector':
            if self.n == 3:
                point = self.regularize(
                    point, point_type=point_type)

                n_points, _ = point.shape

                angle = gs.linalg.norm(point, axis=1)
                angle = gs.expand_dims(angle, axis=1)

                coef_1 = gs.zeros([n_points, 1])
                coef_2 = gs.zeros([n_points, 1])

                # This avoids dividing by 0.
                mask_0 = gs.isclose(angle, 0.)
                mask_0_float = gs.cast(mask_0, gs.float32) + self.epsilon

                coef_1 += mask_0_float * (
                    TAYLOR_COEFFS_1_AT_0[0]
                    + TAYLOR_COEFFS_1_AT_0[2] * angle ** 2
                    + TAYLOR_COEFFS_1_AT_0[4] * angle ** 4
                    + TAYLOR_COEFFS_1_AT_0[6] * angle ** 6)

                coef_2 += mask_0_float * (
                    TAYLOR_COEFFS_2_AT_0[0]
                    + TAYLOR_COEFFS_2_AT_0[2] * angle ** 2
                    + TAYLOR_COEFFS_2_AT_0[4] * angle ** 4
                    + TAYLOR_COEFFS_2_AT_0[6] * angle ** 6)

                # This avoids dividing by 0.
                mask_pi = gs.isclose(angle, gs.pi)
                mask_pi_float = gs.cast(mask_pi, gs.float32) + self.epsilon

                delta_angle = angle - gs.pi
                coef_1 += mask_pi_float * (
                    TAYLOR_COEFFS_1_AT_PI[1] * delta_angle
                    + TAYLOR_COEFFS_1_AT_PI[2] * delta_angle ** 2
                    + TAYLOR_COEFFS_1_AT_PI[3] * delta_angle ** 3
                    + TAYLOR_COEFFS_1_AT_PI[4] * delta_angle ** 4
                    + TAYLOR_COEFFS_1_AT_PI[5] * delta_angle ** 5
                    + TAYLOR_COEFFS_1_AT_PI[6] * delta_angle ** 6)

                angle += mask_0_float
                coef_2 += mask_pi_float * (
                    (1 - coef_1) / angle ** 2)

                # This avoids dividing by 0.
                mask_else = ~mask_0 & ~mask_pi
                mask_else_float = gs.cast(mask_else, gs.float32) + self.epsilon

                # This avoids division by 0.
                angle += mask_pi_float
                coef_1 += mask_else_float * (
                    (angle / 2) / gs.tan(angle / 2))
                coef_2 += mask_else_float * (
                    (1 - coef_1) / angle ** 2)
                jacobian = gs.zeros((n_points, self.dim, self.dim))
                n_points_tensor = gs.array(n_points)
                for i in range(n_points):
                    # This avoids dividing by 0.
                    mask_i_float = (
                        gs.get_mask_i_float(i, n_points_tensor)
                        + self.epsilon)

                    sign = - 1.
                    if left_or_right == 'left':
                        sign = + 1.

                    jacobian_i = (
                        coef_1[i] * gs.eye(self.dim)
                        + coef_2[i] * gs.outer(point[i], point[i])
                        + sign * self.skew_matrix_from_vector(point[i]) / 2.)

                    jacobian += gs.einsum(
                        'n,ij->nij',
                        mask_i_float,
                        jacobian_i)

            else:
                if left_or_right == 'right':
                    raise NotImplementedError(
                        'The jacobian of the right translation'
                        ' is not implemented.')
                jacobian = self.matrix_from_rotation_vector(point)

        elif point_type == 'matrix':
            jacobian = point

        return jacobian

    def random_uniform(self, n_samples=1, point_type=None):
        """Sample in SO(n) with the uniform distribution.

        Parameters
        ----------
        n_samples : int
            the amount of samples
        point_type : str, {'vector', 'matrix'}, optional
            default: self.self.default_point_type

        Returns
        -------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        if point_type is None:
            point_type = self.default_point_type

        random_point = gs.random.rand(n_samples, self.dim) * 2 - 1
        random_point = self.regularize(random_point, point_type='vector')
        if point_type == 'matrix':
            random_point = self.matrix_from_rotation_vector(random_point)

        if n_samples == 1:
            random_point = gs.squeeze(random_point, axis=0)

        return random_point

    @geomstats.vectorization.decorator(['else', 'point', 'point_type'])
    def exp_from_identity(self, tangent_vec, point_type=None):
        """Compute the group exponential of the tangent vector at the identity.

        Parameters
        ----------
        tangent_vec : array-like, shape=[n_samples, {dim, [n, n]}]
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        if point_type == 'matrix' and self.n > 3:
            return gs.linalg.expm(tangent_vec)
        if point_type == 'matrix':
            tangent_vec = self.vector_from_skew_matrix(tangent_vec)
            return self.matrix_from_rotation_vector(tangent_vec)

        return tangent_vec

    @geomstats.vectorization.decorator(['else', 'point', 'point_type'])
    def log_from_identity(self, point, point_type=None):
        """Compute the group logarithm of the point at the identity.

        Parameters
        ----------
        point : array-like, shape=[n_samples, {dim, [n, n]}]
        point_type : str, {'vector', 'matrix'}, optional
            default: self.default_point_type

        Returns
        -------
        tangent_vec : array-like, shape=[n_samples, {dim, [n, n]}]
        """
        if point_type == 'vector':
            tangent_vec = self.regularize(point, point_type=point_type)
        elif point_type == 'matrix':
            point = self.rotation_vector_from_matrix(point)
            tangent_vec = self.skew_matrix_from_vector(point)
        return tangent_vec



"""Unit tests for the Hyperbolic space using Poincaré Ball Model.

We verify poincare ball model by compare results
of squared distance computed with inner_product
(using RiemannianMetric methods) and distance defined
in PoincareBall.
We also verify the distance is the same using differents
coordinates systems.

"""

import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.poincare_ball import PoincareBall


class TestPoincareBallMethods(geomstats.tests.TestCase):
    def setUp(self):
        self.manifold = PoincareBall(2)
        self.metric = self.manifold.metric

        self.hyperboloid_manifold = Hyperboloid(2)
        self.hyperboloid_metric = self.hyperboloid_manifold.metric

    def test_squared_dist(self):
        point_a = gs.array([-0.3, 0.7])
        point_b = gs.array([0.2, 0.5])

        distance_a_b = self.metric.dist(point_a, point_b)
        squared_distance = self.metric.squared_dist(point_a, point_b)

        # TODO(nmiolane):
        # Remove this line when poincare ball is properly vectorized
        squared_distance = gs.to_ndarray(squared_distance, to_ndim=2, axis=1)

        self.assertAllClose(distance_a_b**2, squared_distance, atol=1e-8)

    @geomstats.tests.np_and_pytorch_only
    def test_coordinates(self):
        point_a = gs.array([-0.3, 0.7])
        point_b = gs.array([0.2, 0.5])

        point_a_h =\
            self.manifold.to_coordinates(point_a, 'extrinsic')
        point_b_h =\
            self.manifold.to_coordinates(point_b, 'extrinsic')

        dist_in_ball =\
            self.metric.dist(point_a, point_b)
        dist_in_hype =\
            self.hyperboloid_metric.dist(point_a_h, point_b_h)

        # TODO(ninamiolane): Remove this to_ndarray when poincare_ball
        # is properly vectorized
        dist_in_hype = gs.to_ndarray(dist_in_hype, to_ndim=2, axis=1)

        self.assertAllClose(dist_in_ball, dist_in_hype, atol=1e-8)

    def test_dist_poincare(self):

        point_a = gs.array([0.5, 0.5])
        point_b = gs.array([0.5, -0.5])

        dist_a_b =\
            self.manifold.metric.dist(point_a, point_b)

        result = dist_a_b
        expected = gs.array([[2.887270927429199]])

        self.assertAllClose(result, expected)

    def test_dist_vectorization(self):
        point_a = gs.array([0.2, 0.5])
        point_b = gs.array([[0.3, -0.5], [0.2, 0.2]])

        dist_a_b =\
            self.manifold.metric.dist(point_a, point_b)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.dist(point_a, point_b[i])
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

    def test_mobius_vectorization(self):
        point_a = gs.array([0.5, 0.5])
        point_b = gs.array([[0.5, -0.3], [0.3, 0.4]])

        dist_a_b =\
            self.manifold.metric.mobius_add(point_a, point_b)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.mobius_add(point_a, point_b[i])
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

        dist_a_b =\
            self.manifold.metric.mobius_add(point_b, point_a)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.mobius_add(point_b[i], point_a)
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

    def test_log_vectorization(self):
        point_a = gs.array([0.5, 0.5])
        point_b = gs.array([[0.5, -0.5], [0.4, 0.4]])

        dist_a_b =\
            self.manifold.metric.log(point_a, point_b)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.log(point_a, point_b[i])
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

        dist_a_b =\
            self.manifold.metric.log(point_b, point_a)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.log(point_b[i], point_a)
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

    def test_exp_vectorization(self):
        point_a = gs.array([0.5, 0.5])
        point_b = gs.array([[0.5, -0.5], [0.4, 0.4]])

        dist_a_b =\
            self.manifold.metric.exp(point_a, point_b)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.exp(point_a, point_b[i])
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

        dist_a_b =\
            self.manifold.metric.exp(point_b, point_a)

        result_vect = dist_a_b
        result =\
            [self.manifold.metric.exp(point_b[i], point_a)
             for i in range(len(point_b))]
        result = gs.concatenate(result, axis=0)
        self.assertAllClose(result_vect, result)

    def test_log_poincare(self):

        point = gs.array([[0.3, 0.5]])
        base_point = gs.array([[0.3, 0.3]])

        result = self.manifold.metric.log(point, base_point)
        expected = gs.array([[-0.01733576, 0.21958634]])

        self.manifold.metric.coords_type = 'extrinsic'
        self.assertAllClose(result, expected)

    def test_belong_true_poincare(self):
        point = gs.array([[0.3, 0.5]])
        belong = self.manifold.belongs(point)
        self.assertTrue(belong)

    def test_belong_false_poincare(self):
        point = gs.array([[1.2, 0.5]])
        belong = self.manifold.belongs(point)
        self.assertFalse(belong)

    def test_exp_poincare(self):

        point = gs.array([[0.3, 0.5]])
        base_point = gs.array([[0.3, 0.3]])

        tangent_vec = self.manifold.metric.log(point, base_point)
        result = self.manifold.metric.exp(tangent_vec, base_point)

        self.manifold.metric.coords_type = 'extrinsic'
        self.assertAllClose(result, point)

    def test_ball_retraction(self):
        x = gs.array([[0.5, 0.6], [0.2, -0.1], [0.2, -0.4]])
        y = gs.array([[0.3, 0.5], [0.3, -0.6], [0.3, -0.3]])

        ball_metric = self.manifold.metric
        tangent_vec = ball_metric.log(y, x)
        ball_metric.retraction(tangent_vec, x)



"""Visualization for Geometric Statistics."""

import matplotlib.pyplot as plt

import geomstats.backend as gs
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.hypersphere import Hypersphere
from geomstats.geometry.special_euclidean import SpecialEuclidean
from geomstats.geometry.special_orthogonal import SpecialOrthogonal
from mpl_toolkits.mplot3d import Axes3D  # NOQA

SE3_GROUP = SpecialEuclidean(n=3)
SO3_GROUP = SpecialOrthogonal(n=3)
S1 = Hypersphere(dim=1)
S2 = Hypersphere(dim=2)
H2 = Hyperboloid(dim=2)

AX_SCALE = 1.2

IMPLEMENTED = ['SO3_GROUP', 'SE3_GROUP', 'S1', 'S2',
               'H2_poincare_disk', 'H2_poincare_half_plane', 'H2_klein_disk',
               'poincare_polydisk']


class Arrow3D():
    """An arrow in 3d, i.e. a point and a vector."""

    def __init__(self, point, vector):
        self.point = point
        self.vector = vector

    def draw(self, ax, **quiver_kwargs):
        """Draw the arrow in 3D plot."""
        ax.quiver(self.point[0], self.point[1], self.point[2],
                  self.vector[0], self.vector[1], self.vector[2],
                  **quiver_kwargs)


class Trihedron():
    """A trihedron, i.e. 3 Arrow3Ds at the same point."""

    def __init__(self, point, vec_1, vec_2, vec_3):
        self.arrow_1 = Arrow3D(point, vec_1)
        self.arrow_2 = Arrow3D(point, vec_2)
        self.arrow_3 = Arrow3D(point, vec_3)

    def draw(self, ax, **arrow_draw_kwargs):
        """Draw the trihedron by drawing its 3 Arrow3Ds.

        Arrows are drawn is order using green, red, and blue
        to show the trihedron's orientation.
        """
        if 'color' in arrow_draw_kwargs:
            self.arrow_1.draw(ax, **arrow_draw_kwargs)
            self.arrow_2.draw(ax, **arrow_draw_kwargs)
            self.arrow_3.draw(ax, **arrow_draw_kwargs)
        else:
            blue = '#1f77b4'
            orange = '#ff7f0e'
            green = '#2ca02c'
            self.arrow_1.draw(ax, color=blue, **arrow_draw_kwargs)
            self.arrow_2.draw(ax, color=orange, **arrow_draw_kwargs)
            self.arrow_3.draw(ax, color=green, **arrow_draw_kwargs)


class Circle():
    """Class used to draw a circle."""

    def __init__(self, n_angles=100, points=None):
        angles = gs.linspace(0, 2 * gs.pi, n_angles)
        self.circle_x = gs.cos(angles)
        self.circle_y = gs.sin(angles)
        self.points = []
        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(-ax_s, ax_s),
                 xlabel='X', ylabel='Y')
        return ax

    def add_points(self, points):
        assert gs.all(S1.belongs(points))
        if not isinstance(points, list):
            points = points.tolist()
        self.points.extend(points)

    def draw(self, ax, **plot_kwargs):
        ax.plot(self.circle_x, self.circle_y, color="black")
        if self.points:
            self.draw_points(ax, **plot_kwargs)

    def draw_points(self, ax, points=None, **plot_kwargs):
        if points is None:
            points = self.points
        points = gs.array(points)
        ax.plot(points[:, 0], points[:, 1], marker='o', linestyle="None",
                **plot_kwargs)


class Sphere():
    """Create the arrays sphere_x, sphere_y, sphere_z to plot a sphere.

    Create the arrays sphere_x, sphere_y, sphere_z of values
    to plot the wireframe of a sphere.
    Their shape is (n_meridians, n_circles_latitude).
    """

    def __init__(self, n_meridians=40, n_circles_latitude=None,
                 points=None):
        if n_circles_latitude is None:
            n_circles_latitude = max(n_meridians / 2, 4)

        u, v = gs.meshgrid(
            gs.arange(0, 2 * gs.pi, 2 * gs.pi / n_meridians),
            gs.arange(0, gs.pi, gs.pi / n_circles_latitude))

        self.center = gs.zeros(3)
        self.radius = 1
        self.sphere_x = self.center[0] + self.radius * gs.cos(u) * gs.sin(v)
        self.sphere_y = self.center[1] + self.radius * gs.sin(u) * gs.sin(v)
        self.sphere_z = self.center[2] + self.radius * gs.cos(v)

        self.points = []
        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        if ax is None:
            ax = plt.subplot(111, projection='3d')

        ax_s = AX_SCALE
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(-ax_s, ax_s),
                 zlim=(-ax_s, ax_s),
                 xlabel='X', ylabel='Y', zlabel='Z')
        return ax

    def add_points(self, points):
        assert gs.all(S2.belongs(points))
        if not isinstance(points, list):
            points = points.tolist()
        self.points.extend(points)

    def draw(self, ax, **scatter_kwargs):
        ax.plot_wireframe(self.sphere_x,
                          self.sphere_y,
                          self.sphere_z,
                          color="grey", alpha=0.2)
        if self.points:
            self.draw_points(ax, **scatter_kwargs)

    def draw_points(self, ax, points=None, **scatter_kwargs):
        if points is None:
            points = self.points
        points_x = gs.vstack([point[0] for point in points])
        points_y = gs.vstack([point[1] for point in points])
        points_z = gs.vstack([point[2] for point in points])
        ax.scatter(points_x, points_y, points_z, **scatter_kwargs)

    def fibonnaci_points(self, n_points=16000):
        """Spherical Fibonacci point sets yield nearly uniform point
        distributions on the unit sphere."""
        x_vals = []
        y_vals = []
        z_vals = []

        offset = 2. / n_points
        increment = gs.pi * (3. - gs.sqrt(5.))

        for i in range(n_points):
            y = ((i * offset) - 1) + (offset / 2)
            r = gs.sqrt(1 - pow(y, 2))

            phi = ((i + 1) % n_points) * increment

            x = gs.cos(phi) * r
            z = gs.sin(phi) * r

            x_vals.append(x)
            y_vals.append(y)
            z_vals.append(z)

        x_vals = [(self.radius * i) for i in x_vals]
        y_vals = [(self.radius * i) for i in y_vals]
        z_vals = [(self.radius * i) for i in z_vals]

        return gs.array([x_vals, y_vals, z_vals])

    def plot_heatmap(self, ax,
                     scalar_function,
                     n_points=16000,
                     alpha=0.2,
                     cmap='jet'):
        """Plot a heatmap defined by a loss on the sphere."""
        points = self.fibonnaci_points(n_points)
        intensity = gs.array([scalar_function(x) for x in points.T])
        ax.scatter(points[0, :], points[1, :], points[2, :],
                   c=intensity,
                   alpha=alpha,
                   marker='.',
                   cmap=plt.get_cmap(cmap))


class PoincareDisk():
    def __init__(self, points=None, point_type='extrinsic'):
        self.center = gs.array([0., 0.])
        self.points = []
        self.point_type = point_type

        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(-ax_s, ax_s),
                 xlabel='X', ylabel='Y')
        return ax

    def add_points(self, points):

        if self.point_type == 'extrinsic':
            assert gs.all(H2.belongs(points))
            points = self.convert_to_poincare_coordinates(points)

        if not isinstance(points, list):
            points = points.tolist()

        if gs.all([len(point) == 2 for point in self.points]):
            self.points.extend(points)
        else:
            raise ValueError('Points do not have dimension 2.')

    def convert_to_poincare_coordinates(self, points):
        poincare_coords = points[:, 1:] / (1 + points[:, :1])
        return poincare_coords

    def draw(self, ax, **kwargs):
        circle = plt.Circle((0, 0), radius=1., color='black', fill=False)
        ax.add_artist(circle)
        if len(self.points) > 0:
            if gs.all([len(point) == 2 for point in self.points]):
                points_x = gs.stack(
                    [point[0] for point in self.points], axis=0)
                points_y = gs.stack(
                    [point[1] for point in self.points], axis=0)
                ax.scatter(points_x, points_y, **kwargs)
            else:
                raise ValueError('Points do not have dimension 2.')


class PoincarePolyDisk():
    """Class used to plot points in the Poincare polydisk."""

    def __init__(self, points=None, point_type='ball', n_disks=2):
        self.center = gs.array([0., 0.])
        self.points = []
        self.point_type = point_type
        self.n_disks = n_disks
        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        """Define the ax parameters."""
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(-ax_s, ax_s),
                 xlabel='X', ylabel='Y')
        return ax

    def add_points(self, points):
        """Add points to draw."""
        if self.point_type == 'extrinsic':
            points = self.convert_to_poincare_coordinates(points)
        if not isinstance(points, list):
            points = points.tolist()
        self.points.extend(points)

    def clear_points(self):
        """Clear the points to draw."""
        self.points = []

    def convert_to_poincare_coordinates(self, points):
        """Convert points to poincare coordinates."""
        poincare_coords = points[:, 1:] / (1 + points[:, :1])
        return poincare_coords

    def draw(self, ax, **kwargs):
        """Draw."""
        circle = plt.Circle((0, 0), radius=1., color='black', fill=False)
        ax.add_artist(circle)
        points_x = gs.vstack([point[0] for point in self.points])
        points_y = gs.vstack([point[1] for point in self.points])
        ax.scatter(points_x, points_y, **kwargs)


class PoincareHalfPlane():
    """Class used to plot points in the Poincare Half Plane."""

    def __init__(self, points=None):
        self.points = []
        if points is not None:
            self.add_points(points)

    def add_points(self, points):
        assert gs.all(H2.belongs(points))
        points = self.convert_to_half_plane_coordinates(points)
        if not isinstance(points, list):
            points = points.tolist()
        self.points.extend(points)

    def set_ax(self, ax=None):
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(0., ax_s),
                 xlabel='X', ylabel='Y')
        return ax

    def convert_to_half_plane_coordinates(self, points):
        disk_coords = points[:, 1:] / (1 + points[:, :1])
        disk_x = disk_coords[:, 0]
        disk_y = disk_coords[:, 1]

        half_plane_coords = gs.zeros_like(disk_coords)
        denominator = (disk_x ** 2 + (1 - disk_y) ** 2)
        half_plane_coords[:, 0] = 2 * disk_x / denominator
        half_plane_coords[:, 1] = ((1 - disk_x ** 2 - disk_y ** 2)
                                   / denominator)
        return half_plane_coords

    def draw(self, ax, **kwargs):
        points_x = gs.vstack([point[0] for point in self.points])
        points_y = gs.vstack([point[1] for point in self.points])
        ax.scatter(points_x, points_y, **kwargs)


class KleinDisk():
    def __init__(self, points=None):
        self.center = gs.array([0., 0.])
        self.points = []
        if points is not None:
            self.add_points(points)

    def set_ax(self, ax=None):
        if ax is None:
            ax = plt.subplot()
        ax_s = AX_SCALE
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(-ax_s, ax_s),
                 xlabel='X', ylabel='Y')
        return ax

    def add_points(self, points):
        assert gs.all(H2.belongs(points))
        points = self.convert_to_klein_coordinates(points)
        if not isinstance(points, list):
            points = points.tolist()
        self.points.extend(points)

    def convert_to_klein_coordinates(self, points):
        poincare_coords = points[:, 1:] / (1 + points[:, :1])
        poincare_radius = gs.linalg.norm(
            poincare_coords, axis=1)
        poincare_angle = gs.arctan2(
            poincare_coords[:, 1], poincare_coords[:, 0])

        klein_radius = 2 * poincare_radius / (1 + poincare_radius ** 2)
        klein_angle = poincare_angle

        klein_coords = gs.zeros_like(poincare_coords)
        klein_coords[:, 0] = klein_radius * gs.cos(klein_angle)
        klein_coords[:, 1] = klein_radius * gs.sin(klein_angle)
        return klein_coords

    def draw(self, ax, **kwargs):
        circle = plt.Circle((0, 0), radius=1., color='black', fill=False)
        ax.add_artist(circle)
        points_x = gs.vstack([point[0] for point in self.points])
        points_y = gs.vstack([point[1] for point in self.points])
        ax.scatter(points_x, points_y, **kwargs)


def convert_to_trihedron(point, space=None):
    """Transform a rigid point into a trihedron.

    Transform a rigid point
    into a trihedron s.t.:
    - the trihedron's base point is the translation of the origin
    of R^3 by the translation part of point,
    - the trihedron's orientation is the rotation of the canonical basis
    of R^3 by the rotation part of point.
    """
    point = gs.to_ndarray(point, to_ndim=2)
    n_points, _ = point.shape

    dim_rotations = SO3_GROUP.dim

    if space == 'SE3_GROUP':
        rot_vec = point[:, :dim_rotations]
        translation = point[:, dim_rotations:]
    elif space == 'SO3_GROUP':
        rot_vec = point
        translation = gs.zeros((n_points, 3))
    else:
        raise NotImplementedError(
            'Trihedrons are only implemented for SO(3) and SE(3).')

    rot_mat = SO3_GROUP.matrix_from_rotation_vector(rot_vec)
    rot_mat = SO3_GROUP.projection(rot_mat)
    basis_vec_1 = gs.array([1, 0, 0])
    basis_vec_2 = gs.array([0, 1, 0])
    basis_vec_3 = gs.array([0, 0, 1])

    trihedrons = []
    for i in range(n_points):
        trihedron_vec_1 = gs.dot(rot_mat[i], basis_vec_1)
        trihedron_vec_2 = gs.dot(rot_mat[i], basis_vec_2)
        trihedron_vec_3 = gs.dot(rot_mat[i], basis_vec_3)
        trihedron = Trihedron(translation[i],
                              trihedron_vec_1,
                              trihedron_vec_2,
                              trihedron_vec_3)
        trihedrons.append(trihedron)
    return trihedrons


def plot(points, ax=None, space=None,
         point_type='extrinsic', **point_draw_kwargs):
    """Plot points in the 3D Special Euclidean Group.

    Plot points in the 3D Special Euclidean Group,
    by showing them as trihedrons.
    """
    if space not in IMPLEMENTED:
        raise NotImplementedError(
            'The plot function is not implemented'
            ' for space {}. The spaces available for visualization'
            ' are: {}.'.format(space, IMPLEMENTED))

    if points is None:
        raise ValueError("No points given for plotting.")

    points = gs.to_ndarray(points, to_ndim=2)

    if space in ('SO3_GROUP', 'SE3_GROUP'):
        if ax is None:
            ax = plt.subplot(111, projection='3d')
        if space == 'SE3_GROUP':
            ax_s = AX_SCALE * gs.amax(gs.abs(points[:, 3:6]))
        elif space == 'SO3_GROUP':
            ax_s = AX_SCALE * gs.amax(gs.abs(points[:, :3]))
        plt.setp(ax,
                 xlim=(-ax_s, ax_s),
                 ylim=(-ax_s, ax_s),
                 zlim=(-ax_s, ax_s),
                 xlabel='X', ylabel='Y', zlabel='Z')
        trihedrons = convert_to_trihedron(points, space=space)
        for t in trihedrons:
            t.draw(ax, **point_draw_kwargs)

    elif space == 'S1':
        circle = Circle()
        ax = circle.set_ax(ax=ax)
        circle.add_points(points)
        circle.draw(ax, **point_draw_kwargs)

    elif space == 'S2':
        sphere = Sphere()
        ax = sphere.set_ax(ax=ax)
        sphere.add_points(points)
        sphere.draw(ax, **point_draw_kwargs)

    elif space == 'H2_poincare_disk':
        poincare_disk = PoincareDisk(point_type=point_type)
        ax = poincare_disk.set_ax(ax=ax)
        poincare_disk.add_points(points)
        poincare_disk.draw(ax, **point_draw_kwargs)

    elif space == 'poincare_polydisk':
        n_disks = points.shape[1]
        poincare_poly_disk = PoincarePolyDisk(point_type=point_type,
                                              n_disks=n_disks)
        n_columns = gs.ceil(n_disks ** 0.5)
        n_rows = gs.ceil(n_disks / n_columns)

        axis_list = []

        for i_disk in range(n_disks):
            axis_list.append(ax.add_subplot(n_rows, n_columns, i_disk + 1))

        for i_disk, ax in enumerate(axis_list):
            ax = poincare_poly_disk.set_ax(ax=ax)
            poincare_poly_disk.clear_points()
            poincare_poly_disk.add_points(points[:, i_disk, ...])
            poincare_poly_disk.draw(ax, **point_draw_kwargs)

    elif space == 'H2_poincare_half_plane':
        poincare_half_plane = PoincareHalfPlane()
        ax = poincare_half_plane.set_ax(ax=ax)
        poincare_half_plane.add_points(points)
        poincare_half_plane.draw(ax, **point_draw_kwargs)

    elif space == 'H2_klein_disk':
        klein_disk = KleinDisk()
        ax = klein_disk.set_ax(ax=ax)
        klein_disk.add_points(points)
        klein_disk.draw(ax, **point_draw_kwargs)

    return ax



"""Statistical Manifold of beta distributions with the Fisher metric."""

from autograd.scipy.special import polygamma
from scipy.integrate import odeint
from scipy.integrate import solve_bvp
from scipy.stats import beta

import geomstats.backend as gs
import geomstats.error
from geomstats.geometry.embedded_manifold import EmbeddedManifold
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric

N_STEPS = 100
EPSILON = 1e-6


class BetaDistributions(EmbeddedManifold):
    r"""Class for the manifold of beta distributions.

    This is :math: Beta = `R_+^* \times R_+^*`.
    """

    def __init__(self):
        super(BetaDistributions, self).__init__(
            dim=2, embedding_manifold=Euclidean(dim=2))

    def belongs(self, point, point_type=None):
        """Evaluate if a point belongs to the manifold of beta distributions.

        The statistical manifold of beta distributions is the upper right
        quadrant of the euclidean 2-plane.

        Parameters
        ----------
        point : array-like, shape=[n_samples, 2]
            the point of which to check whether it belongs to Beta

        Returns
        -------
        belongs : array-like, shape=[n_samples, 1]
            array of booleans indicating whether point belongs the Beta
        """
        point = gs.to_ndarray(point, to_ndim=2)
        n_points, point_dim = point.shape
        belongs = point_dim == self.dim
        belongs = gs.to_ndarray(belongs, to_ndim=1)
        belongs = gs.tile(belongs, n_points)
        belongs = belongs * gs.greater(point, 0).all(axis=1)
        return belongs[0] if n_points == 1 else belongs

    @staticmethod
    def random_uniform(n_samples=1, bound=10.0):
        """Sample parameters of beta distributions.

        The uniform distribution on [0, bound]^2 is used.

        Parameters
        ----------
        n_samples : int, optional
        bound : float, optional
            scalar to scale samples

        Returns
        -------
        samples : array-like, shape=[n_samples, 2]
        """
        return bound * gs.random.rand(n_samples, 2)

    def sample(self, point, n_samples=1):
        """Sample from the beta distribution.

        Sample from the beta distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape [n_points, 2]
        n_samples : int
            the number of points to sample with each pair of parameter in
            point

        Returns
        -------
        samples : array-like, shape=[n_points, n_samples]
        """
        point = gs.to_ndarray(point, to_ndim=2)
        geomstats.error.check_belongs(
            point, self, 'beta_distributions')
        samples = []
        for param_a, param_b in point:
            samples.append(beta.rvs(param_a, param_b, size=n_samples))
        return samples[0] if len(point) == 1 else gs.stack(samples)

    def maximum_likelihood_fit(self, data, loc=0, scale=1):
        """Estimate parameters from samples.

        This a wrapper around scipy's maximum likelihood estimator to
        estimate the parameters of a beta distribution from samples.

        Parameters
        ----------
        data : array-like, shape=[n_distributions, n_samples]
            the data to estimate parameters from. Arrays of
            different length may be passed.
        loc : float, optional
            the location parameter of the distribution to estimate parameters
            from. It is kept fixed during optimization
            default: 0
        scale : float, optional
            the scale parameter of the distribution to estimate parameters
            from. It is kept fixed during optimization
            default: 1
        Returns
        -------
        parameter : array-like, shape=[n_samples, 2]
        """
        data = gs.to_ndarray(
            gs.where(data == 1., 1 - EPSILON, data), to_ndim=2)
        parameters = []
        for sample in data:
            param_a, param_b, _, _ = beta.fit(sample, floc=loc, fscale=scale)
            parameters.append(gs.array([param_a, param_b]))
        return parameters[0] if len(data) == 1 else gs.stack(parameters)


class BetaMetric(RiemannianMetric):
    """Class for the Fisher information metric on beta distributions."""

    def __init__(self):
        super(RiemannianMetric, self).__init__(dim=2)

    @staticmethod
    def metric_det(param_a, param_b):
        """Compute the determinant of the metric.

        Parameters
        ----------
        param_a : array-like, shape=[n_samples,]
        param_b : array-like, shape=[n_samples,]

        Returns
        -------
        metric_det : array-like, shape=[n_samples,]
        """
        metric_det = polygamma(1, param_a) * polygamma(1, param_b) - \
            polygamma(1, param_a + param_b) * (polygamma(1, param_a) +
                                               polygamma(1, param_b))
        return metric_det

    def inner_product_matrix(self, base_point=None):
        """Compute inner product matrix at the tangent space at base point.

        Parameters
        ----------
        base_point : array-like, shape=[n_samples, 2]

        Returns
        -------
        base_point : array-like, shape=[n_samples, 2, 2]
        """
        if base_point is not None:
            raise ValueError('The metric depends on the base point.')
        base_point = gs.to_ndarray(base_point, to_ndim=2)
        matrices = []
        for point in base_point:
            param_a, param_b = point
            g0 = gs.array(
                [polygamma(1, param_a) - polygamma(1, param_a + param_b),
                 - polygamma(1, param_a + param_b)])
            g1 = gs.array(
                [- polygamma(1, param_a + param_b), polygamma(1, param_b)
                 - polygamma(1, param_a + param_b)])
            matrices.append(gs.stack([g0, g1]))
        return gs.stack(matrices)

    def christoffels(self, base_point):
        """Compute the Christoffel symbols.

        Compute the Christoffel symbols of the Fisher information metric on
        Beta.

        Parameters
        ----------
        base_point : array-like, shape=[n_samples, 2]

        Returns
        -------
        christoffels : array-like, shape=[n_samples, 2, 2, 2]
        """

        def coefficients(param_a, param_b):
            metric_det = self.metric_det(param_a, param_b)
            c1 = (polygamma(2, param_a) * polygamma(1, param_b) -
                  polygamma(2, param_a) * polygamma(1, param_a + param_b) -
                  polygamma(1, param_b) * polygamma(2, param_a + param_b)) / (
                2 * metric_det)
            c2 = - polygamma(1, param_b) * polygamma(2, param_a + param_b) / (
                2 * metric_det)
            c3 = (polygamma(2, param_b) * polygamma(1, param_a + param_b) -
                  polygamma(1, param_b) * polygamma(2, param_a + param_b)) / (
                2 * metric_det)
            return c1, c2, c3

        if base_point is None:
            raise ValueError('Christoffels require a base point.')
        base_point = gs.to_ndarray(base_point, to_ndim=2)
        param_a, param_b = base_point[:, 0], base_point[:, 1]
        c1, c2, c3 = coefficients(param_a, param_b)
        c4, c5, c6 = coefficients(param_b, param_a)
        christoffel = []
        for d1, d2, d3, d4, d5, d6 in zip(c1, c2, c3, c4, c5, c6):
            gamma_0 = gs.array([[d1, d2], [d2, d3]])
            gamma_1 = gs.array([[d6, d5], [d5, d4]])
            christoffel.append(gs.stack([gamma_0, gamma_1]))
        if len(base_point) == 1:
            return christoffel[0]
        return gs.stack(christoffel)

    def exp(self, tangent_vec, base_point, n_steps=N_STEPS):
        """Exponential map associated to the Fisher information metric.

        Exponential map at base_point of tangent_vec computed by integration
        of the geodesic equation (initial value problem), using the
        christoffel symbols.

        Parameters
        ----------
        tangent_vec : array-like, shape=[n_samples, dim]
        base_point : array-like, shape=[n_samples, dim]
        n_steps : int

        Returns
        -------
        exp : array-like, shape=[n_samples, dim]
        """
        base_point = gs.to_ndarray(base_point, to_ndim=2)
        tangent_vec = gs.to_ndarray(tangent_vec, to_ndim=2)

        def ivp(state, time):
            """Reformat the initial value problem geodesic ODE."""
            position, velocity = state[:2], state[2:]
            eq = self.geodesic_equation(velocity=velocity, position=position)
            return gs.hstack([velocity, eq])

        times = gs.linspace(0, 1, n_steps + 1)
        exp = []
        for point, vec in zip(base_point, tangent_vec):
            initial_state = gs.hstack([point, vec])
            geodesic = odeint(
                ivp, initial_state, times, tuple(), rtol=1e-6)
            exp.append(geodesic[-1, :2])
        return exp[0] if len(base_point) == 1 else gs.stack(exp)

    def log(self, point, base_point, n_steps=N_STEPS):
        """Compute logarithm map associated to the Fisher information metric.

        Solve the boundary value problem associated to the geodesic ordinary
        differential equation (ODE) using the Christoffel symbols.

        Parameters
        ----------
        point : array-like, shape=[n_samples, dim]
        base_point : array-like, shape=[n_samples, dim]
        n_steps : int

        Returns
        -------
        tangent_vec : array-like, shape=[n_samples, dim]
            the initial velocity of the geodesic starting at base_point and
            reaching point at time 1
        """
        stop_time = 1.
        t = gs.linspace(0, stop_time, n_steps)
        point = gs.to_ndarray(point, to_ndim=2)
        base_point = gs.to_ndarray(base_point, to_ndim=2)

        def initialize(end_point, start_point):
            a0, b0 = start_point
            a1, b1 = end_point
            lin_init = gs.zeros([2 * self.dim, n_steps])
            lin_init[0, :] = gs.linspace(a0, a1, n_steps)
            lin_init[1, :] = gs.linspace(b0, b1, n_steps)
            lin_init[2, :-1] = (lin_init[0, 1:] - lin_init[0, :-1]) * n_steps
            lin_init[3, :-1] = (lin_init[1, 1:] - lin_init[1, :-1]) * n_steps
            lin_init[2, -1] = lin_init[2, -2]
            lin_init[3, -1] = lin_init[3, -2]
            return lin_init

        def bvp(time, state):
            """Reformat the boundary value problem geodesic ODE.

            Parameters
            ----------
                state :  vector of the state variables: y = [a,b,u,v]
                time :  time
            """
            position, velocity = state[:2].T, state[2:].T
            eq = self.geodesic_equation(
                velocity=velocity, position=position)
            return gs.vstack((velocity.T, eq.T))

        def boundary_cond(
                state_a, state_b, point_0_a, point_0_b, point_1_a, point_1_b):
            return gs.array(
                [state_a[0] - point_0_a,
                 state_a[1] - point_0_b,
                 state_b[0] - point_1_a,
                 state_b[1] - point_1_b])

        log = []
        for bp, pt in zip(base_point, point):
            geodesic_init = initialize(pt, bp)
            base_point_a, base_point_b = bp
            point_a, point_b = pt

            def bc(y0, y1):
                return boundary_cond(
                    y0, y1, base_point_a, base_point_b, point_a, point_b)

            solution = solve_bvp(bvp, bc, t, geodesic_init)
            geodesic = solution.sol(t)
            geodesic = geodesic[:2, :]
            log.append(n_steps * (geodesic[:, 1] - geodesic[:, 0]))

        return log[0] if len(base_point) == 1 else gs.stack(log)



"""Unit tests for the Poincare Polydisk."""


import geomstats.backend as gs
import geomstats.tests
from geomstats.geometry.hyperboloid import Hyperboloid
from geomstats.geometry.poincare_polydisk import PoincarePolydisk


class TestPoincarePolydiskMethods(geomstats.tests.TestCase):
    """Class defining the Poincare polydisk tests."""

    def setUp(self):
        """Define the elements to test."""
        gs.random.seed(1234)
        self.n_disks = 5
        self.space = PoincarePolydisk(n_disks=self.n_disks)
        self.metric = self.space.metric

    def test_dimension(self):
        """Test the dimension."""
        expected = self.n_disks * 2
        result = self.space.dim
        self.assertAllClose(result, expected)

    def test_metric_signature(self):
        """Test the signature."""
        expected = (self.n_disks * 2, 0, 0)
        result = self.metric.signature
        self.assertAllClose(result, expected)

    @geomstats.tests.np_and_pytorch_only
    def test_product_distance_extrinsic_representation(self):
        """Test the distance using the extrinsic representation."""
        coords_type = 'extrinsic'
        point_a_intrinsic = gs.array([0.01, 0.0])
        point_b_intrinsic = gs.array([0.0, 0.0])
        hyperbolic_space = Hyperboloid(dim=2)
        point_a = hyperbolic_space.from_coordinates(
            point_a_intrinsic, 'intrinsic')
        point_b = hyperbolic_space.from_coordinates(
            point_b_intrinsic, 'intrinsic')

        duplicate_point_a = gs.vstack([point_a, point_a])
        duplicate_point_b = gs.vstack([point_b, point_b])

        single_disk = PoincarePolydisk(n_disks=1, coords_type=coords_type)
        two_disks = PoincarePolydisk(n_disks=2, coords_type=coords_type)

        distance_single_disk = single_disk.metric.dist(point_a, point_b)
        distance_two_disks = two_disks.metric.dist(
            duplicate_point_a, duplicate_point_b)
        result = distance_two_disks
        expected = 3 ** 0.5 * distance_single_disk
        self.assertAllClose(result, expected)



"""Online kmeans algorithm on Manifolds."""

import logging

from sklearn.base import BaseEstimator, ClusterMixin

import geomstats.backend as gs


# TODO(nkoep): Move this into the OnlineKMeans class.

def online_kmeans(X, metric, n_clusters, n_repetitions=20,
                  tolerance=1e-5, max_iter=5e4):
    """Perform online K-means clustering.

    Perform online version of k-means algorithm on data contained in X.
    The data points are treated sequentially and the cluster centers are
    updated one at a time. This version of k-means avoids computing the
    mean of each cluster at each iteration and is therefore less
    computationally intensive than the offline version.

    In the setting of quantization of probability distributions, this
    algorithm is also known as Competitive Learning Riemannian Quantization.
    It computes the closest approximation of the empirical distribution of
    data by a discrete distribution supported by a smaller number of points
    with respect to the Wasserstein distance. This smaller number of points
    is n_clusters.

    Parameters
    ----------
    X : array-like, shape=[n_samples, n_features]
        Input data. It is treated sequentially by the algorithm, i.e.
        one datum is chosen randomly at each iteration.
    metric : object
        Metric of the space in which the data lives. At each iteration,
        one of the cluster centers is moved in the direction of the new
        datum, according the exponential map of the underlying space, which
        is a method of metric.
    n_clusters : int
        Number of clusters of the k-means clustering, or number of desired
        atoms of the quantized distribution.
    n_repetitions : int, default=20
        The cluster centers are updated using decreasing step sizes, each
        of which stays constant for n_repetitions iterations to allow a better
        exploration of the data points.
    max_iter : int, default=5e4
        Maximum number of iterations. If it is reached, the
        quantization may be inacurate.

    Returns
    -------
    cluster_centers : array, shape=[n_clusters, n_features]
        Coordinates of cluster centers.
    labels : array, shape=[n_samples]
        Cluster labels for each point.
    """
    n_samples = X.shape[0]

    random_indices = gs.random.randint(low=0, high=n_samples,
                                       size=(n_clusters,))
    cluster_centers = gs.get_slice(X, gs.cast(random_indices, gs.int32))

    gap = 1.0
    iteration = 0

    while iteration < max_iter:
        iteration += 1
        step_size = gs.floor(gs.array(iteration / n_repetitions)) + 1

        random_index = gs.random.randint(low=0, high=n_samples, size=(1,))
        point = gs.get_slice(X, gs.cast(random_index, gs.int32))

        index_to_update = metric.closest_neighbor_index(point, cluster_centers)
        center_to_update = gs.copy(
            gs.get_slice(cluster_centers, index_to_update))

        tangent_vec_update = metric.log(
            point=point, base_point=center_to_update
        ) / (step_size + 1)
        new_center = metric.exp(
            tangent_vec=tangent_vec_update,
            base_point=center_to_update
        )
        gap = metric.dist(center_to_update, new_center)
        if gap == 0 and iteration == 1:
            gap = gs.array(1.0)

        cluster_centers[index_to_update, :] = new_center

        if gs.isclose(gap, 0, atol=tolerance):
            break

    if iteration == max_iter - 1:
        logging.warning(
            'Maximum number of iterations {} reached. The'
            'clustering may be inaccurate'.format(max_iter))

    labels = gs.zeros(n_samples)
    for i in range(n_samples):
        labels[i] = int(metric.closest_neighbor_index(X[i], cluster_centers))

    return cluster_centers, labels


class OnlineKMeans(BaseEstimator, ClusterMixin):
    """Online k-means clustering.

    Online k-means clustering seeks to divide a set of data points into
    a specified number of classes, while minimizing intra-class variance.
    It is closely linked to discrete quantization, which computes the closest
    approximation of the empirical distribution of the dataset by a discrete
    distribution supported by a smaller number of points with respect to the
    Wasserstein distance. The algorithm used can either be seen as an online
    version of the k-means algorithm or as Competitive Learning Riemannian
    Quantization (see [LBP2019]_).

    Parameters
    ----------
    metric : object
        Metric of the space in which the data lives. At each iteration,
        one of the cluster centers is moved in the direction of the new
        datum, according the exponential map of the underlying space, which
        is a method of metric.
    n_clusters : int
        Number of clusters of the k-means clustering, or number of desired
        atoms of the quantized distribution.
    n_repetitions : int, default=20
        The cluster centers are updated using decreasing step sizes, each
        of which stays constant for n_repetitions iterations to allow a better
        exploration of the data points.
    max_iter : int, default=5e4
        Maximum number of iterations. If it is reached, the
        quantization may be inacurate.

    Attributes
    ----------
    cluster_centers_ : array, [n_clusters, n_features]
        Coordinates of cluster centers.
    labels_ :
        Labels of each point.

    Example
    -------
    >>> from geomstats.geometry.hypersphere import Hypersphere
    >>> from geomstats.learning.onlinekmeans import OnlineKmeans
    >>> sphere = Hypersphere(dim=2)
    >>> metric = sphere.metric
    >>> X = sphere.random_von_mises_fisher(kappa=10, n_samples=50)
    >>> clustering = OnlineKmeans(metric=metric,n_clusters=4).fit(X)
    >>> clustering.cluster_centers_
    >>> clustering.labels_

    References
    ----------
    .. [LBP2019] A. Le Brigant and S. Puechmorel, Optimal Riemannian
       quantization with an application to air traffic analysis. J. Multivar.
       Anal. 173 (2019), 685 - 703.
    """

    def __init__(self, metric, n_clusters, n_repetitions=20,
                 tolerance=1e-5, max_iter=5e4, point_type='vector'):
        self.metric = metric
        self.n_clusters = n_clusters
        self.n_repetitions = n_repetitions
        self.tolerance = tolerance
        self.max_iter = max_iter
        self.point_type = point_type

    def fit(self, X):
        """Perform clustering.

        Parameters
        ----------
        X : array-like, shape=[n_features, n_samples]
            Samples to cluster.
        """
        self.cluster_centers_, self.labels_ = \
            online_kmeans(X=X, metric=self.metric,
                          n_clusters=self.n_clusters,
                          n_repetitions=self.n_repetitions,
                          tolerance=self.tolerance,
                          max_iter=self.max_iter)

        return self

    def predict(self, point):
        """Predict the closest cluster each sample in X belongs to.

        Parameters
        ----------
        X : array-like, shape=[n_features]
            New data to predict.

        Returns
        -------
        labels : Index of the cluster each sample belongs to.
        """
        return self.metric.closest_neighbor_index(point, self.cluster_centers_)



"""
Testing class for geomstats.

This class abstracts the backend type.
"""

import os
import unittest

import numpy as np

import geomstats.backend as gs


def pytorch_backend():
    """Check if pytorch is set as backend."""
    return os.environ['GEOMSTATS_BACKEND'] == 'pytorch'


def tf_backend():
    """Check if tensorflow is set as backend."""
    return os.environ['GEOMSTATS_BACKEND'] == 'tensorflow'


def np_backend():
    """Check if numpy is set as backend."""
    return os.environ['GEOMSTATS_BACKEND'] == 'numpy'


def np_only(test_item):
    """Decorate to filter tests for numpy only."""
    if np_backend():
        return test_item
    return unittest.skip('Test for numpy backend only.')(test_item)


def pytorch_only(test_item):
    """Decorate to filter tests for pytorch only."""
    if pytorch_backend():
        return test_item
    return unittest.skip('Test for pytorch backend only.')(test_item)


def tf_only(test_item):
    """Decorate to filter tests for tensorflow only."""
    if tf_backend():
        return test_item
    return unittest.skip('Test for tensorflow backend only.')(test_item)


def np_and_tf_only(test_item):
    """Decorate to filter tests for numpy and tensorflow only."""
    if np_backend() or tf_backend():
        return test_item
    return unittest.skip('Test for numpy and tensorflow backends only.')(
        test_item)


def np_and_pytorch_only(test_item):
    """Decorate to filter tests for numpy and pytorch only."""
    if np_backend() or pytorch_backend():
        return test_item
    return unittest.skip('Test for numpy and pytorch backends only.')(
        test_item)


_TestBaseClass = unittest.TestCase
if tf_backend():
    import tensorflow as tf
    _TestBaseClass = tf.test.TestCase


class TestCase(_TestBaseClass):
    def assertAllClose(self, a, b, rtol=1e-6, atol=1e-6):
        if tf_backend():
            return super().assertAllClose(a, b, rtol=rtol, atol=atol)
        if np_backend():
            return np.testing.assert_allclose(a, b, rtol=rtol, atol=atol)
        return self.assertTrue(gs.allclose(a, b, rtol=rtol, atol=atol))

    def assertAllCloseToNp(self, a, np_a, rtol=1e-6, atol=1e-6):
        are_same_shape = np.all(a.shape == np_a.shape)
        if pytorch_backend():
            are_same = np.all(np.array(a) == np_a)
        else:
            are_same = np.all(a == np_a)
        return are_same_shape and are_same

    def assertShapeEqual(self, a, b):
        if tf_backend():
            return super().assertShapeEqual(a, b)
        super().assertEqual(a.shape, b.shape)

    @classmethod
    def setUpClass(cls):
        if tf_backend():
            os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'



"""Decorator to handle vectorization.

This abstracts the backend type.
"""

import geomstats.backend as gs

POINT_TYPES = ['scalar', 'vector', 'matrix']
FLEXIBLE_TYPE = 'point'
OTHER_TYPES = ['point_type', 'else']

POINT_TYPES_TO_NDIMS = {
    'scalar': 2,
    'vector': 2,
    'matrix': 3}

ERROR_MSG = 'Invalid type: %s.'


def get_n_points(points, point_type):
    """Compute the number of points.

    Parameters
    ----------
    points : array-like
        Input points.
    point_type : str, {'scalar', ''vector', 'matrix'}
        Type of input points.

    Returns
    -------
    n_points : int
        Number of points.
    """
    n_points = 1

    points_ndim = gs.ndim(points)

    is_vect_scalar = (point_type == 'scalar' and points_ndim == 1)
    is_vect_vector = (point_type == 'vector' and points_ndim == 2)
    is_vect_matrix = (point_type == 'matrix' and points_ndim == 3)

    if is_vect_scalar or is_vect_vector or is_vect_matrix:
        n_points = gs.shape(points)[0]
    return n_points


def decorator(input_types):
    """Vectorize geomstats functions.

    This decorator assumes that its function:
    - works with fully-vectorized inputs,
    - returns fully-vectorized outputs,

    where "fully-vectorized" means that:
    - one scalar has shape [1, 1],
    - n scalars have shape [n, 1],
    - one d-D vector has shape [1, d],
    - n d-D vectors have shape [n, d],etc
    etc.

    The decorator:
    - gets the types of all inputs of its function:
        - args,
        - kwargs,
        - optional kwargs,
            - e.g. input_type=None,
    - gets the type of the output of its function,
        - e.g. distinguishes between 1D 'vector' vs 'scalar',
    - gets the initial shapes of all inputs of its function,
    - if needed, adapts the types of the inputs,
        - e.g. distinguishes between 'vector' or 'matrix' inputs,
        using variables 'point_type' or 'default_point_type',
    - converts the inputs into fully-vectorized inputs,
    - calls the function,
    - adapts the output shapes to match the users' expectations,
    using the initial shapes of the inputs.

    Parameters
    ----------
    input_types : list
        List of inputs' input_types, including for optional inputs.
        The `input_type`s of optional inputs will not be read
        by the decorator if the corresponding input is not given.
    """
    if not isinstance(input_types, list):
        input_types = list(input_types)

    def aux_decorator(function):
        def wrapper(*args, **kwargs):
            args_types, kwargs_types, opt_kwargs_types, is_scal = get_types(
                input_types, args, kwargs)
            args_types, kwargs_types, kwargs = adapt_types(
                args_types, kwargs_types, opt_kwargs_types, args, kwargs)
            args_kwargs_types = args_types + kwargs_types

            args_shapes = get_initial_shapes(args_types, args)
            kwargs_shapes = get_initial_shapes(kwargs_types, kwargs.values())
            initial_shapes = args_shapes + kwargs_shapes

            vect_args = vectorize_args(args_types, args)
            vect_kwargs = vectorize_kwargs(kwargs_types, kwargs)

            result = function(*vect_args, **vect_kwargs)

            result = adapt_result(
                result, initial_shapes, args_kwargs_types, is_scal)

            return result
        return wrapper
    return aux_decorator


def get_types(input_types, args, kwargs):
    """Extract the types of args, kwargs, optional kwargs and output.

    Parameters
    ----------
    input_types : list
        List of inputs' input_types, including for optional inputs.
    args : tuple
        Args of a function.
    kwargs : dict
        Kwargs of a function.

    Returns
    -------
    args_types : list
        Types of args.
    kwargs_types : list
        Types of kwargs.
    opt_kwargs_types : list
        Types of optional kwargs.
    is_scal : bool
        Boolean determining if the output is a scalar.
    """
    len_args = len(args)
    len_kwargs = len(kwargs)
    len_total = len_args + len_kwargs

    args_types = input_types[:len_args]
    kwargs_types = input_types[len_args:len_total]

    opt_kwargs_types = []
    is_scal = True
    if len(input_types) > len_total:
        opt_kwargs_types = input_types[len_total:]
        last_input_type = input_types[-1]
        if 'output_' in last_input_type:
            if last_input_type != 'output_scalar':
                is_scal = False
                opt_kwargs_types = input_types[len_total:-1]
    return (args_types, kwargs_types, opt_kwargs_types, is_scal)


def adapt_types(
        args_types, kwargs_types,
        opt_kwargs_types, args, kwargs):
    """Adapt the list of input input_types.

    Some functions are implemented with array-like arguments that can be either
    'vector' or 'matrix' depending on the value of the 'point_type'
    argument.

    This function reads the 'point_type' argument, and adapt the actual
    type of the input array-like arguments.

    Parameters
    ----------
    args_types : list
        Types of args.
    kwargs_types : list
        Types of kwargs.
    opt_kwargs_types : list
        Types of optional kwargs.
    args : tuple
        Args of a function.
    kwargs : dict
        Kwargs of a function.

    Returns
    -------
    args_types : list
        Adapted types of args.
    kwargs_types : list
        Adapted types of kwargs.
    """
    in_args = 'point_type' in args_types
    in_kwargs = 'point_type' in kwargs_types
    in_optional = 'point_type' in opt_kwargs_types

    if in_args or in_kwargs or in_optional:
        if in_args:
            i_input_type = args_types.index('point_type')
            input_type = args[i_input_type]
        elif in_kwargs:
            input_type = kwargs['point_type']

        elif in_optional:
            obj = args[0]
            input_type = obj.default_point_type
            kwargs['point_type'] = input_type
            kwargs_types.append('point_type')

        args_types = [
            input_type if pt == FLEXIBLE_TYPE else pt
            for pt in args_types]
        kwargs_types = [
            input_type if pt == FLEXIBLE_TYPE else pt
            for pt in kwargs_types]
    return args_types, kwargs_types, kwargs


def get_initial_shapes(input_types, args):
    """Extract shapes and ndims of input args or kwargs values.

    Store the shapes of the input args, or kwargs values,
    that are array-like, store None otherwise.

    Parameters
    ----------
    input_types : list
        Point types corresponding to the args, or kwargs values.
    args : tuple or dict_values
        Args, or kwargs values, of a function.

    Returns
    -------
    in_shapes : list
        Shapes of array-like input args, or kwargs values.
    """
    in_shapes = []

    for arg, input_type in zip(args, input_types):
        if input_type == 'scalar':
            arg = gs.array(arg)

        if input_type in POINT_TYPES and arg is not None:
            in_shapes.append(gs.shape(arg))
        elif input_type in OTHER_TYPES or arg is None:
            in_shapes.append(None)
        else:
            raise ValueError(ERROR_MSG % input_type)
    return in_shapes


def vectorize_args(input_types, args):
    """Vectorize input args.

    Transform input array-like args into their fully-vectorized form,
    where "fully-vectorized" means that:
    - one scalar has shape [1, 1],
    - n scalars have shape [n, 1],
    - one d-D vector has shape [1, d],
    - n d-D vectors have shape [n, d],
    etc.

    Parameters
    ----------
    input_types : list
        Point types corresponding to the args.
    args : tuple
        Args of a function.

    Returns
    -------
    vect_args : tuple
        Args of the function in their fully-vectorized form.
    """
    vect_args = []
    for arg, input_type in zip(args, input_types):
        if input_type == 'scalar':
            vect_arg = gs.to_ndarray(arg, to_ndim=1)
            vect_arg = gs.to_ndarray(vect_arg, to_ndim=2, axis=1)
        elif input_type in POINT_TYPES and arg is not None:
            vect_arg = gs.to_ndarray(
                arg, to_ndim=POINT_TYPES_TO_NDIMS[input_type])
        elif input_type in OTHER_TYPES or arg is None:
            vect_arg = arg
        else:
            raise ValueError(ERROR_MSG % input_type)
        vect_args.append(vect_arg)
    return tuple(vect_args)


def vectorize_kwargs(input_types, kwargs):
    """Vectorize input kwargs.

    Transform input array-like kwargs into their fully-vectorized form,
    where "fully-vectorized" means that:
    - one scalar has shape [1, 1],
    - n scalars have shape [n, 1],
    - one d-D vector has shape [1, d],
    - n d-D vectors have shape [n, d],
    etc.

    Parameters
    ----------
    input_types :list
        Point types corresponding to the args.
    kwargs : dict
        Kwargs of a function.

    Returns
    -------
    vect_kwargs : dict
        Kwargs of the function in their fully-vectorized form.
    """
    vect_kwargs = {}
    for key_arg, input_type in zip(kwargs.keys(), input_types):
        arg = kwargs[key_arg]
        if input_type == 'scalar':
            vect_arg = gs.to_ndarray(arg, to_ndim=1)
            vect_arg = gs.to_ndarray(vect_arg, to_ndim=2, axis=1)
        elif input_type in POINT_TYPES and arg is not None:
            vect_arg = gs.to_ndarray(
                arg, to_ndim=POINT_TYPES_TO_NDIMS[input_type])
        elif input_type in OTHER_TYPES or arg is None:
            vect_arg = arg
        else:
            raise ValueError(ERROR_MSG % input_type)
        vect_kwargs[key_arg] = vect_arg
    return vect_kwargs


def adapt_result(result, initial_shapes, args_kwargs_types, is_scal):
    """Adapt shape of output.

    This function squeezes the dim 0 or 1 of the output, depending on:
    - the type of the output: scalar vs else,
    - the initial shapes or args and kwargs provided by the user.

    Parameters
    ----------
    result : unspecified
        Output of the function.
    initial_shapes : list
        Shapes of args and kwargs provided by the user.
    args_kwargs_types : list
        Types of args and kwargs.
    is_scal : bool
        Boolean determining if the output 'result' is a scalar.

    Returns
    -------
    result : unspecified
        Output of the function, with adapted shape.
    """
    if squeeze_output_dim_1(
            result, initial_shapes, args_kwargs_types, is_scal):
        if result.shape[1] == 1:
            result = gs.squeeze(result, axis=1)

    if squeeze_output_dim_0(result, initial_shapes, args_kwargs_types):
        if result.shape[0] == 1:
            result = gs.squeeze(result, axis=0)

    return result


def squeeze_output_dim_0(result, in_shapes, input_types):
    """Determine if the output needs to be squeezed on dim 0.

    The dimension 0 is squeezed iff all input parameters:
    - contain one sample,
    - have the corresponding dimension 0 squeezed,
    i.e. if all input parameters have ndim strictly less than the ndim
    corresponding to their vectorized shape.

    Parameters
    ----------
    in_ndims : list
        Initial ndims of input parameters, as entered by the user.
    input_types : list
        Associated list of input_type of input parameters.

    Returns
    -------
    squeeze : bool
        Boolean deciding whether to squeeze dim 0 of the output.
    """
    if isinstance(result, tuple):
        return False
    if isinstance(result, list):
        return False

    for in_shape, input_type in zip(in_shapes, input_types):
        if input_type not in POINT_TYPES:
            continue
        in_ndim = None
        if in_shape is not None:
            in_ndim = len(in_shape)
        if in_ndim is not None:
            vect_ndim = POINT_TYPES_TO_NDIMS[input_type]
            if in_ndim > vect_ndim:
                raise ValueError(
                    'Fully-vectorizing an input can only increase its ndim.')
            if in_ndim == vect_ndim:
                return False
    return True


def squeeze_output_dim_1(result, in_shapes, input_types, is_scal=True):
    """Determine if the output needs to be squeezed on dim 1.

    This happens if the user represents scalars as array of shapes:
    [n_samples,] instead of [n_samples, 1]
    Dimension 1 is squeezed by default if input_type is 'scalar'.
    Dimension 1 is not squeezed if the user inputs at least one scalar with
    a singleton in dimension 1.

    Parameters
    ----------
    result: array-like
        Result output by the function, before reshaping.
    in_shapes : list
        Initial shapes of input parameters, as entered by the user.
    input_types : list
        Associated list of input_type of input parameters.

    Returns
    -------
    squeeze : bool
        Boolean deciding whether to squeeze dim 1 of the output.
    """
    if not is_scal:
        return False
    if not is_scalar(result):
        return False

    for shape, input_type in zip(in_shapes, input_types):
        if input_type == 'scalar':
            ndim = len(shape)
            if ndim > 2:
                raise ValueError('The ndim of a scalar cannot be > 2.')
            if ndim == 2:
                return False
    return True


def is_scalar(vect_array):
    """Test if a "fully-vectorized" array represents a scalar.

    Parameters
    ----------
    vect_array :  array-like
        Array to be tested.

    Returns
    -------
    is_scalar : bool
        Boolean determining if vect_array is a fully-vectorized scalar.
    """
    if isinstance(vect_array, tuple):
        return False
    if isinstance(vect_array, list):
        return False

    has_ndim_2 = vect_array.ndim == 2
    if not has_ndim_2:
        return False
    has_singleton_dim_1 = vect_array.shape[1] == 1
    return has_singleton_dim_1



"""Manifold for sets of landmarks that belong to any given manifold."""

import math

import geomstats.backend as gs
import geomstats.vectorization
from geomstats.geometry.manifold import Manifold
from geomstats.geometry.riemannian_metric import RiemannianMetric


class Landmarks(Manifold):
    """Class for landmarks."""

    def __init__(self, ambient_manifold, n_landmarks=None):
        """Construct an instance of the LandmarksSpace class.

        Parameters
        ----------
        ambient_manifold : object from the class Manifold
        n_landmarks
        """
        dimension = None
        if n_landmarks:
            self.dim = n_landmarks * ambient_manifold.dim
        super(Landmarks, self).__init__(dim=dimension)
        self.ambient_manifold = ambient_manifold
        self.l2_metric = L2Metric(self.ambient_manifold)
        self.n_landmarks = n_landmarks

    def belongs(self, point):
        """Compute whether or not a point belongs to the manifold.

        Parameters
        ----------
        point

        Returns
        -------
        belongs : bool
        """
        # TODO(ninamiolane): vectorize this method
        belongs = self.ambient_manifold.belongs(point)
        belongs = gs.all(belongs)
        return belongs


class L2Metric(RiemannianMetric):
    """L2 Riemannian metric on the space of landmarks."""

    def __init__(self, ambient_manifold):
        super(L2Metric, self).__init__(
            dim=math.inf,
            signature=(math.inf, 0, 0))
        self.ambient_manifold = ambient_manifold
        self.ambient_metric = ambient_manifold.metric

    @geomstats.vectorization.decorator(['else', 'matrix', 'matrix', 'matrix'])
    def inner_product(self, tangent_vec_a, tangent_vec_b, base_landmarks):
        """Compute inner product between tangent vectors at base landmark set.

        Parameters
        ----------
        tangent_vec_a
        tangent_vec_b
        base_landmarks

        Returns
        -------
        inner_prod
        """
        if not (tangent_vec_a.shape == tangent_vec_b.shape
                and tangent_vec_a.shape == base_landmarks.shape):
            raise NotImplementedError

        n_landmark_sets, n_landmarks_per_set, n_coords = tangent_vec_a.shape

        new_dim = n_landmark_sets * n_landmarks_per_set
        tangent_vec_a = gs.reshape(tangent_vec_a, (new_dim, n_coords))
        tangent_vec_b = gs.reshape(tangent_vec_b, (new_dim, n_coords))
        base_landmarks = gs.reshape(base_landmarks, (new_dim, n_coords))

        inner_prod = self.ambient_metric.inner_product(
            tangent_vec_a, tangent_vec_b, base_landmarks)
        inner_prod = gs.reshape(
            inner_prod, (n_landmark_sets, n_landmarks_per_set))
        inner_prod = gs.sum(inner_prod, -1)

        n_landmarks_per_set_float = gs.array(n_landmarks_per_set)
        n_landmarks_per_set_float = gs.cast(
            n_landmarks_per_set_float, gs.float32)
        inner_prod = inner_prod / n_landmarks_per_set_float
        inner_prod = gs.to_ndarray(inner_prod, to_ndim=1)
        inner_prod = gs.to_ndarray(inner_prod, to_ndim=2, axis=1)

        return inner_prod

    @geomstats.vectorization.decorator(['else', 'matrix', 'matrix'])
    def dist(self, landmarks_a, landmarks_b):
        """Compute geodesic distance between two landmark sets.

        Parameters
        ----------
        landmarks_a
        landmarks_b

        Returns
        -------
        dist
        """
        if landmarks_a.shape != landmarks_b.shape:
            raise NotImplementedError

        n_landmark_sets, n_landmarks_per_set, n_coords = landmarks_a.shape

        landmarks_a = gs.reshape(
            landmarks_a, (n_landmark_sets * n_landmarks_per_set, n_coords))
        landmarks_b = gs.reshape(
            landmarks_b, (n_landmark_sets * n_landmarks_per_set, n_coords))

        dist = self.ambient_metric.dist(landmarks_a, landmarks_b)
        dist = gs.reshape(dist, (n_landmark_sets, n_landmarks_per_set))
        n_landmarks_per_set_float = gs.array(n_landmarks_per_set)
        n_landmarks_per_set_float = gs.cast(
            n_landmarks_per_set_float, gs.float32)
        dist = gs.sqrt(gs.sum(dist ** 2, -1) / n_landmarks_per_set_float)
        dist = gs.to_ndarray(dist, to_ndim=1)
        dist = gs.to_ndarray(dist, to_ndim=2, axis=1)

        return dist

    @geomstats.vectorization.decorator(['else', 'matrix', 'matrix'])
    def exp(self, tangent_vec, base_landmarks):
        """Compute Riemannian exponential of tan vector wrt base landmark set.

        Parameters
        ----------
        tangent_vec
        base_landmarks

        Returns
        -------
        exp
        """
        n_landmark_sets, n_landmarks_per_set, n_coords = base_landmarks.shape
        n_tangent_vecs = tangent_vec.shape[0]

        new_dim = n_landmark_sets * n_landmarks_per_set
        new_base_landmarks = gs.reshape(base_landmarks, (new_dim, n_coords))
        new_tangent_vec = gs.reshape(tangent_vec, (new_dim, n_coords))

        exp = self.ambient_metric.exp(new_tangent_vec, new_base_landmarks)
        exp = gs.reshape(exp, (n_tangent_vecs, n_landmarks_per_set, n_coords))
        exp = gs.squeeze(exp)

        return exp

    @geomstats.vectorization.decorator(['else', 'matrix', 'matrix'])
    def log(self, landmarks, base_landmarks):
        """Compute Riemannian log of a set of landmarks wrt base landmark set.

        Parameters
        ----------
        landmarks
        base_landmarks

        Returns
        -------
        log
        """
        if landmarks.shape != base_landmarks.shape:
            raise NotImplementedError

        n_landmark_sets, n_landmarks_per_set, n_coords = landmarks.shape

        landmarks = gs.reshape(
            landmarks, (n_landmark_sets * n_landmarks_per_set, n_coords))
        base_landmarks = gs.reshape(
            base_landmarks, (n_landmark_sets * n_landmarks_per_set, n_coords))
        log = self.ambient_metric.log(landmarks, base_landmarks)
        log = gs.reshape(log, (n_landmark_sets, n_landmarks_per_set, n_coords))
        log = gs.squeeze(log)

        return log

    def geodesic(self, initial_landmarks,
                 end_landmarks=None, initial_tangent_vec=None):
        """Compute geodesic from initial & end landmark set (or init. tan vec).

        Compute the geodesic specified either by an initial landmark set and
        an end landmark set, or by an initial landmark set and an initial
        tangent vector.

        Parameters
        ----------
        initial_landmarks
        end_landmarks
        initial_tangent_vec

        Returns
        -------
        landmarks_on_geodesic
        """
        landmarks_ndim = 2
        initial_landmarks = gs.to_ndarray(
            initial_landmarks, to_ndim=landmarks_ndim + 1)

        if end_landmarks is None and initial_tangent_vec is None:
            raise ValueError(
                'Specify an end landmark set or an initial tangent'
                'vector to define the geodesic.')
        if end_landmarks is not None:
            end_landmarks = gs.to_ndarray(
                end_landmarks, to_ndim=landmarks_ndim + 1)
            shooting_tangent_vec = self.log(landmarks=end_landmarks,
                                            base_landmarks=initial_landmarks)
            if initial_tangent_vec is not None:
                if not gs.allclose(shooting_tangent_vec, initial_tangent_vec):
                    raise RuntimeError(
                        'The shooting tangent vector is too'
                        ' far from the initial tangent vector.')
            initial_tangent_vec = shooting_tangent_vec
        initial_tangent_vec = gs.array(initial_tangent_vec)
        initial_tangent_vec = gs.to_ndarray(initial_tangent_vec,
                                            to_ndim=landmarks_ndim + 1)

        def landmarks_on_geodesic(t):
            t = gs.cast(t, gs.float32)
            t = gs.to_ndarray(t, to_ndim=1)
            t = gs.to_ndarray(t, to_ndim=2, axis=1)
            new_initial_landmarks = gs.to_ndarray(
                initial_landmarks, to_ndim=landmarks_ndim + 1)
            new_initial_tangent_vec = gs.to_ndarray(
                initial_tangent_vec, to_ndim=landmarks_ndim + 1)

            tangent_vecs = gs.einsum('il,nkm->ikm', t, new_initial_tangent_vec)

            def point_on_landmarks(tangent_vec):
                if gs.ndim(tangent_vec) < 2:
                    raise RuntimeError
                exp = self.exp(
                    tangent_vec=tangent_vec,
                    base_landmarks=new_initial_landmarks)
                return exp

            landmarks_at_time_t = gs.vectorize(
                tangent_vecs,
                point_on_landmarks,
                signature='(i,j)->(i,j)')

            return landmarks_at_time_t

        return landmarks_on_geodesic



"""The KNN classifier on manifolds."""

from sklearn.neighbors import KNeighborsClassifier


class KNearestNeighborsClassifier(KNeighborsClassifier):
    """Classifier implementing the k-nearest neighbors vote on manifolds.

    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default.
    weights : string or callable, optional (default = 'uniform')
        Weight function used in prediction. Possible values:
        - 'uniform' : uniform weights.  All points in each neighborhood
          are weighted equally.
        - 'distance' : weight points by the inverse of their distance.
          in this case, closer neighbors of a query point will have a
          greater influence than neighbors which are further away.
        - [callable] : a user-defined function which accepts an
          array of distances, and returns an array of the same shape
          containing the weights.
    p : integer, optional (default = 2)
        Power parameter for the 'minkowski' string distance.
        When p = 1, this is equivalent to using manhattan_distance (l1),
        and euclidean_distance (l2) for p = 2.
        For arbitrary p, minkowski_distance (l_p) is used.
    distance : string or callable, optional (default = 'minkowski')
        The distance metric to use.
        The default distance is minkowski, and with p=2 is equivalent to the
        standard Euclidean distance.
        See the documentation of the DistanceMetric class in the scikit-learn
        library for a list of available distances.
        If distance is "precomputed", X is assumed to be a distance matrix and
        must be square during fit.
    distance_params : dict, optional (default = None)
        Additional keyword arguments for the distance function.
    n_jobs : int or None, optional (default = None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1; ``-1`` means using all processors.

    Attributes
    ----------
    classes_ : array, shape=[n_classes,]
        Class labels known to the classifier
    effective_metric_ : string or callable
        The distance metric used. It will be same as the `distance` parameter
        or a synonym of it, e.g. 'euclidean' if the `distance` parameter set to
        'minkowski' and `p` parameter set to 2.
    effective_metric_params_ : dict
        Additional keyword arguments for the distance function.
        For most distances will be same with `distance_params` parameter,
        but may also contain the `p` parameter value if the
        `effective_metric_` attribute is set to 'minkowski'.
    outputs_2d_ : bool
        False when `y`'s shape is (n_samples, ) or (n_samples, 1) during fit
        otherwise True.

    References
    ----------
    This algorithm uses the scikit-learn library:
    https://github.com/scikit-learn/scikit-learn/blob/95d4f0841/sklearn/
    neighbors/_classification.py#L25
    """

    def __init__(self, n_neighbors=5,
                 weights='uniform',
                 p=2,
                 distance='minkowski',
                 distance_params=None,
                 n_jobs=None,
                 **kwargs):

        super().__init__(
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm='brute',
            p=p,
            metric=distance,
            metric_params=distance_params,
            n_jobs=n_jobs,
            **kwargs)
